const P = require('bluebird');
const _chunk = require('lodash/chunk');
const noop = require('lodash/noop');
const pickBy = require('lodash/pickBy');

const genBatchOp =
  (batchOp) =>
  (givenItems, opts = {}) => {
    const { batchSize = 100, concurrency = 1, batchCb = noop, dryrun = false, dryrunCb, runCloudFunctions } = opts;

    if (!givenItems || !givenItems.length) return P.resolve();

    const items = givenItems.filter(Boolean);

    if (!items.length) return P.resolve();

    if (!items.every((i) => i.docRef)) {
      // TODO: check if docRef is really a firestore ref
      return P.reject(new Error('items must at least have a docRef property'));
    }

    // eslint-disable-next-line prefer-destructuring
    const firestore = items[0].docRef.firestore;

    return P.resolve(_chunk(items, batchSize)).then((chunks) =>
      P.map(
        chunks,
        (chunk, batchNr) => {
          const batch = firestore.batch();

          if (dryrun && !dryrunCb) return undefined;

          if (dryrunCb) {
            return dryrunCb(
              chunk.map((c) =>
                pickBy({
                  op: c.op || c.operator || batchOp,
                  docRef: c.docRef.path,
                  docData: c.docData,
                  docOptions: c.docOptions,
                })
              ),
              batchNr + 1,
              chunks.length
            );
          }

          chunk.forEach((c) => {
            const op = c.op || c.operator || batchOp;
            const docData = c.docData || {};
            // set migration state in order to prevent data mutation on CF processing
            // this field will be deleted by CF
            if (Object.keys(docData).length && runCloudFunctions === false) {
              docData._migration = true;
            }

            if (!['set', 'update', 'delete'].includes(op)) {
              throw new Error(`${op} is not a valid batch operator, try one of (set, update, delete)`);
            }

            if (op === 'set') batch.set(c.docRef, docData, c.docOptions || {});

            if (op === 'update') batch.update(c.docRef, docData); // docData is now updateMap

            if (op === 'delete') batch.delete(c.docRef);
          });

          return batch.commit().then(() => batchCb(chunk, batchNr + 1, chunks.length));
        },
        { concurrency }
      )
    );
  };

const batchSet = genBatchOp('set');
const batchDelete = genBatchOp('delete');
const batchUpdate = genBatchOp('update');
const batch = genBatchOp();

module.exports = {
  batchSet,
  batchDelete,
  batchUpdate,
  batch,
};
