#!/usr/bin/env node

const P = require('bluebird');
const ajvErrors = require('better-ajv-errors');
const chalk = require('chalk');

const { genValidator, stringifyErrors } = require('../../validate');
const { stringIsDoc, stringIsCollection, findCollectionType, findDocType } = require('../../match');
const { genIteratorFuncs } = require('../../iterate');
const { dataToPojo } = require('../..');

const { genFirestore } = require('../helpers');

const {
  verbose,
  throw: doThrow,
  _: args,
} = require('yargs')
  .alias(`v`, `verbose`)
  .describe(`v`, `Increase verbosity`)

  .alias(`t`, `throw`)
  .describe(`t`, `Halt on error`)

  .command('firestorePath', 'either a document or a collection')
  .demandCommand(1, 'Please specify a firestore path').argv;
const fsPaths = args.filter((arg) => {
  if (stringIsDoc(arg) && findDocType(arg)) return true;

  if (stringIsCollection(arg) && findCollectionType(arg)) return true;

  // console.log(`invalid firestore path ${arg}`); //eslint-disable-line
});

const firestore = genFirestore();
const validator = genValidator();

const { iterateCollection } = genIteratorFuncs(firestore);

const validateOpts = {
  debug: false,
  doThrow: true,
};

let failed = 0;

const validateDoc = (doc, isCol) => {
  process.stdout.write(`validating  ${doc.ref.firestore.projectId} ${doc.ref.path}`); //eslint-disable-line
  try {
    validator.validateDoc(doc, validateOpts);
    process.stdout.write(chalk.green(' [OK] \n'));
  } catch (e) {
    if (e instanceof validator.SchemaValidationError) {
      failed++;
      process.stdout.write(chalk.red(' [ERROR] \n'));

      if (!isCol || (isCol && verbose)) {
        console.log(ajvErrors(e.schema, dataToPojo(e.data, firestore), e.schemaErrors, { indent: 2 })); //eslint-disable-line
        console.log(stringifyErrors(e.schemaErrors)); //eslint-disable-line
      }
    }

    if (doThrow) throw e;
  }
};

const qOpts = {
  concurrency: 25,
  pageSize: 250,
};

P.each(fsPaths, (fsPath) => {
  if (stringIsDoc(fsPath))
    return firestore
      .doc(fsPath)
      .get()
      .then((doc) => {
        if (doc.exists) validateDoc(doc);
      });

  return iterateCollection(fsPath, (d) => validateDoc(d, true), qOpts).then((d) => {
    console.log(`validated ${d.length} docs from ${fsPath} (failed: ${failed})`); //eslint-disable-line
    failed = 0;
  });
});
