#!/usr/bin/env node
const ajvErrors = require('better-ajv-errors');
const chalk = require('chalk');

const { genValidator } = require('../../validate');
const { patterns, urlPatternCollectionParts } = require('../../match');
const { genIteratorFuncs } = require('../../iterate');
const { dataToPojo } = require('../..');

const { genFirestore } = require('../helpers');

const keys = require('lodash/keys');
const reduce = require('lodash/reduce');

const {
  verbose,
  throw: doThrow,
  _: args,
} = require('yargs')
  .alias(`v`, `verbose`)
  .describe(`v`, `Increase verbosity`)
  // .option         (`v`, { default: false })

  .alias(`t`, `throw`)
  .describe(`t`, `Halt on error`)
  // .option         (`v`, { default: false })

  .command('firestorePattern', 'firestore path pattern')
  .demandCommand(1, 'Please specify a firestore path pattern').argv;
const type = args[0];

if (!keys(patterns).includes(type)) throw new Error(`no firestore match pattern found for: ${type}`);

const collectionParts = urlPatternCollectionParts(patterns[type]);

if (!(collectionParts.length === 1 || collectionParts.length === 2))
  throw new Error(`collection for ${type} is deeper than 2 collections`);

const firestore = genFirestore();
const validator = genValidator();
const { iterateSubCollection, iterateCollection } = genIteratorFuncs(firestore);

const isSubCol = collectionParts.length === 2;

const qOpts = {
  concurrency: 25,
  pageSize: 250,
};

const validateOpts = {
  debug: false,
  doThrow: true,
};

let failed = 0;

const validateDoc = (doc) => {
  process.stdout.write(`validating  ${doc.ref.firestore.projectId} ${doc.ref.path}`); //eslint-disable-line
  try {
    validator.validateDoc(doc, validateOpts);
    process.stdout.write(chalk.green(' [OK] \n'));
  } catch (e) {
    if (e instanceof validator.SchemaValidationError) {
      failed++;
      process.stdout.write(chalk.red(' [ERROR] \n'));

      if (verbose) {
        const ferrs = ajvErrors(e.schema, dataToPojo(e.data, firestore), e.schemaErrors, { indent: 2 });
        console.log(ferrs || stringifyErrors(e.schemaErrors)); //eslint-disable-line
      }
    }

    if (doThrow) throw e;
  }
};

if (isSubCol) {
  iterateSubCollection(collectionParts[0], collectionParts[1], validateDoc, qOpts, qOpts).then((d) => {
    const colIdsCount = keys(d).length;
    const subColIdsCount = reduce(d, (acc, val) => (acc += val.length), 0);
    return console.log(
      `validated ${subColIdsCount} ${collectionParts[1]} in ${colIdsCount} ${collectionParts[0]} (${failed} failed)`
    ); //eslint-disable-line
  });
} else {
  iterateCollection(collectionParts[0], validateDoc, qOpts).then((d) => {
    console.log(`validated ${d.length} ${collectionParts[0]} (${failed} failed) `); //eslint-disable-line
  });
}
