#!/usr/bin/env node
const yargs = require('yargs');
const progressStream = require('progress-stream');

const createLogger = require('@hatsa/log');

const { genFirestore } = require('../helpers');
const genBatch = require('../../stream/batch');
const { batch: fsBatch } = require('../../batch');
const { genIteratorFuncs } = require('../../iterate');
const firecodeMigrator = require('./firecode');

const logger = createLogger();

const batchSize = 500;

const migrate = async ({ name, firestore, dryrun }) => {
  let migratorSettings;
  try {
    // eslint-disable-next-line global-require, import/no-dynamic-require
    migratorSettings = require(`./migrators/${name}`)(firestore);
  } catch (e) {
    throw new Error(`${name} migrator not found`);
  }

  if (migratorSettings.runWithFirecode) {
    try {
      await firecodeMigrator({ batchSize, logger, migratorSettings, dryrun });
    } catch (e) {
      logger.error('Firecode migration error ', e);
    }
    return null;
  }

  const str = progressStream({ time: 5000, speed: 5000, objectMode: true });
  str.on('progress', (progress) => logger.info('progress', progress));

  const batchHandlerOpts = {
    batchSize,
    handleBatch: (batch) =>
      fsBatch(batch.map(migratorSettings.handle), {
        batchSize,
        dryrunCb:
          dryrun &&
          ((b) => {
            logger.info(`[DRYRUN] processed ${b.length}`);
            // logger.info('batch', b.map(i => i.docRef));
            // logger.info('batch', b);
          }),
        runCloudFunctions: migratorSettings.runCloudFunctions,
      }),
    cb: (batchStats) => logger.info('batchStats', batchStats),
  };

  if (migratorSettings.id) batchHandlerOpts.id = migratorSettings.id;

  if (migratorSettings.merge) batchHandlerOpts.merge = migratorSettings.merge;

  if (migratorSettings.filter) batchHandlerOpts.filter = migratorSettings.filter;

  const { streamCollection } = genIteratorFuncs(firestore);
  const query = migratorSettings.query.limit(migratorSettings.query._queryOptions.limit || 1000);
  const docStream = streamCollection(null, null, { query, logger });

  docStream.pipe(str).pipe(genBatch(batchHandlerOpts));
  return null;
};

const { nodryrun, _: args } = yargs
  .describe('nodryrun', 'No dryrun')
  .command('migrator', 'migrator name')
  .demandCommand(1, 'Please specify a migrator').argv;

const name = args[0];
const dryrun = !nodryrun;

(async () => {
  try {
    await migrate({
      name,
      firestore: genFirestore(),
      dryrun,
    });
  } catch (e) {
    logger.error('Migrated with errors. ', e);
  }
})();
