// const P = require('bluebird');
const omit = require('lodash/omit');

const { patterns, urlPatternCollectionParts } = require('../match');
const { genIteratorFuncs } = require('../iterate');

const { genValidator, SchemaValidationError } = require('../validate');

const { batchSet, batchDelete } = require('../batch');

const { sha1 } = require('@hatsa/utils/crypto');

const MAX_BUFFER_LENGTH = 5;

const collectErrors = (firestore) => (type) => {
  const now = new Date();

  if (!Object.keys(patterns).includes(type)) throw new Error(`no firestore match pattern found for: ${type}`);

  const patternType = patterns[type];
  const colGroup = patternType.colgroup;
  const collectionParts = urlPatternCollectionParts(patternType);

  if (!colGroup && !(collectionParts.length === 1 || collectionParts.length === 2))
    throw new Error(`collection for ${type} is deeper than 2 collections and does not have a collectionGroup defined`);

  const { iterateSubCollection, iterateCollection, iterateCollectionGroup } = genIteratorFuncs(firestore);

  const { validateDoc } = genValidator();

  let failedValidations = [];
  const stats = {
    now,
    type,
    total: 0,
    failed: 0,
    fixed: 0,
  };

  const flush = () => {
    // console.log(`flushing ${failedValidations.length} items`); //eslint-disable-line

    const batch = failedValidations.map((fv) => ({
      docRef: firestore.doc(`schema-errors/${sha1(fv.docRef.path)}`),
      docData: {
        updated_at: now,
        type: fv.type,
        docRef: fv.docRef,
        schemaErrors: fv.schemaErrors,
        errorHash: sha1(JSON.stringify(fv.schemaErrors)),
      },
    }));

    return batchSet(batch).then(() => (failedValidations = []));
  };

  const handleFailedValidation = (item) => {
    failedValidations.push(item);
    stats.failed += 1;

    console.log(item.docRef.path); //eslint-disable-line

    if (failedValidations.length < MAX_BUFFER_LENGTH) return;

    return flush();
  };

  const handleDocValidation = (doc) => {
    try {
      stats.total += 1;
      validateDoc(doc, { debug: false, doThrow: true });
    } catch (e) {
      if (e instanceof SchemaValidationError) {
        return handleFailedValidation({
          type: type,
          docRef: doc.ref,
          schemaErrors: (e.schemaErrors || []).map((se) => omit(se, 'data', 'schema', 'parentSchema')),
        });
      } else throw e;
    }
  };

  const removeFixedErrors = () =>
    iterateCollection('schema-errors', null, {
      pageSize: 500,
      select: ['updated_at', 'type'],
      orderBy: 'updated_at',
      where: [
        ['type', '==', type],
        ['updated_at', '<', now],
      ],
    })
      .then((docs) => docs.map((doc) => ({ docRef: doc.ref })))
      .tap((refsToDelete) => {
        stats.fixed = refsToDelete.length || 0;
      })
      .then((refsToDelete) => batchDelete(refsToDelete, { debug: false }));
  const logStats = () => console.log({ stats }); //eslint-disable-line

  let prom;

  const qOpts = {
    pageSize: 500,
    concurrency: 1,
  };

  if (colGroup) prom = iterateCollectionGroup(colGroup, handleDocValidation, qOpts);

  if (!prom && collectionParts.length === 1) prom = iterateCollection(collectionParts[0], handleDocValidation, qOpts);

  if (!prom) {
    const col = collectionParts[0];
    const subcol = collectionParts[1];
    prom = iterateSubCollection(col, subcol, handleDocValidation, qOpts, qOpts);
  }

  return Promise.resolve(prom).then(flush).then(removeFixedErrors).then(logStats).catch(console.error); //eslint-disable-line
};

module.exports = collectErrors;
