const P = require('bluebird');
const fs = require('fs');
const { batchUpdate } = require('../../../batch');

/**
 * Comment out the following lines to prevent profiles from being update:
 * https://github.com/hatsa-com/hatsa/blob/master/packages/cloud-functions/firestore/boards/onUpdate.js#L63
 * https://github.com/hatsa-com/hatsa/blob/master/packages/cloud-functions/firestore/boards/items/onWrite.js#L102
 */

module.exports = async ({ admin, bigQuery, dryrun, config, logger }) => {
  const FirestoreBatchSize = 500;
  const BigQueryLimit = 5000;

  const firestore = admin.firestore();
  const startTime = process.hrtime.bigint();
  const getExecutionTime = () => `${Number(process.hrtime.bigint() - startTime) / 1e9} seconds`;

  const stats = {
    total: 0,
    touched: 0,
    not_exist: 0,
  };
  const gcp = config.get('google_cloud_project') || 'hatsa-tst-01';

  const fetchBoardsItems = async ({ page: _page, onPage }) => {
    const limit = BigQueryLimit;
    const page = _page || 1;
    const offset = limit * page - limit;

    logger.info(`[${gcp}] Query board items from big query - ${limit}/${offset}`);

    const query = fs
      .readFileSync(`${__dirname}/query.sql`)
      .toString()
      .replace(/@gcp/g, gcp)
      .replace('@limit', limit)
      .replace('@offset', offset);

    const [rows] = await bigQuery.query({ query, location: 'US' });

    stats.total += rows.length;
    await onPage(rows);

    if (rows.length === 0 || rows.length < limit) {
      return null;
    }

    return fetchBoardsItems({ page: page + 1, onPage });
  };

  const batchCb = (chunk, batchNr, totalBatches) =>
    logger.info(`updated ${chunk.length} (batch ${batchNr}/${totalBatches})`);

  const dryrunCb = dryrun && batchCb;

  const recordsToDocRefs = async (boardItems) => {
    const docRefs = await P.map(
      boardItems,
      async (docData) => {
        if (dryrun) {
          logger.debug(`[update] - boards/${docData.boardId}/items/${docData.itemId}`);
        }

        const docPath = `boards/${docData.boardId}/items/${docData.itemId}`;
        const docSnap = await firestore.doc(docPath).get();

        if (!docSnap.exists) {
          stats.not_exist += 1;
          return null;
        }

        return {
          docRef: docSnap.ref,
          docData: {
            touch: true,
            urlref: admin.firestore.FieldValue.delete(),
          },
        };
      },
      { concurrency: FirestoreBatchSize }
    );

    return docRefs.filter((update) => {
      if (update) {
        stats.touched += 1;
        return true;
      }

      return false;
    });
  };

  await fetchBoardsItems({
    onPage: async (boardItems) =>
      batchUpdate(await recordsToDocRefs(boardItems), {
        batchSize: FirestoreBatchSize,
        concurrency: FirestoreBatchSize,
        batchCb,
        dryrun,
        dryrunCb,
      }),
  });

  logger.info('stats', { ...stats, execution_time: getExecutionTime() });
};
