/* eslint-disable vars-on-top, import/no-extraneous-dependencies */
const P = require('bluebird');
const csv = require('csv-string');
const fs = require('fs');

/*
 The goal of the fixer is to fill in the "_copernica_id" prop in the firestore profiles.
 We read the content of "cp_users.csv" (the backup of the copernica profiles).
 We fetch profiles without "_copernica_id" (the list is fetched from profiles synced to BigQuery).
 For each profile without "_copernica_id," we try to find a corresponding record in the copernica backup file.
 If we found such a record - set the "_copernica_id" prop of the firestore profile.
*/
module.exports = async ({ firestore, bigQuery, dryrun, admin, config, logger }) => {
  const stats = {
    profiles_total: 0,
    profiles_touched: 0,
    profiles_ignored: 0,
    copernica_profiles_total: 0,
    copernica_profiles_not_found: 0,
  };
  const gcp = config.get('google_cloud_project') || 'hatsa-tst-01';

  const fetchProfiles = async () => {
    logger.info(`[${gcp}] Query profiles from big query`);
    const query = fs
      .readFileSync(`${__dirname}/profiles-query.sql`)
      .toString()
      .replace('@tableName', `${gcp}.firestore.profiles_raw_latest`);
    const [result] = await bigQuery.query({
      query,
      location: 'US',
    });

    return result;
  };

  const fetchCopernicaProfiles = async () => {
    const bucket = `gs://${gcp}.appspot.com`;
    const fileName = 'cp_users.csv';

    logger.info(`Reading copernica profiles from - "${bucket}/${fileName}"`);

    const fileContent = await admin.app().storage().bucket(bucket).file(fileName).download();
    const result = csv.parse(fileContent.toString(), { output: 'objects' });

    stats.copernica_profiles_total = result.length;
    return result;
  };

  const profilesToFix = await fetchProfiles();
  const copernicaProfiles = await fetchCopernicaProfiles();

  const handler = async ({ profileId }) => {
    stats.profiles_total += 1;

    const copernicaProfile = copernicaProfiles.find((p) => p.uid === profileId);

    if (!copernicaProfile) {
      logger.info(`[${profileId}] copernica profile is not found`);
      stats.copernica_profiles_not_found += 1;
      return;
    }

    const profileRef = firestore.doc(`profiles/${profileId}`);
    const profileDoc = await profileRef.get();

    if (!profileDoc.exists) {
      if (dryrun) logger.info(`[${profileId}] profile does not exist`);
      return;
    }

    const profileCopernicaId = profileDoc.get('_copernica_id');

    if (profileCopernicaId && profileCopernicaId === copernicaProfile.id) {
      stats.profiles_ignored += 1;
      return;
    }

    stats.profiles_touched += 1;

    if (!dryrun) {
      profileRef.update({ _copernica_id: copernicaProfile.id });
    } else {
      logger.info(`[${profileId}] updating "_copernica_id" to ${copernicaProfile.id}`);
    }
  };

  await P.map(profilesToFix, handler, { concurrency: 10 });
  logger.info('stats', stats);
};
