const { DELETED_AT_PROP } = require('@hatsa/constants');
const { genIteratorFuncs } = require('../../iterate');

const LinesSeparator = '\n';
const ColumnsSeparator = ',';

const Column = {
  ID: 0,
  UID: 9,
  Email: 6,
  DisplayName: 16,
  IsAdmin: 14,
};

module.exports = async ({ config, admin, firestore, dryrun, logger }) => {
  const { iterateCollection } = genIteratorFuncs(firestore);
  const gcp = config.get('google_cloud_project') || 'hatsa-tst-01';
  const bucket = `gs://${gcp}.appspot.com`;
  const fileName = 'cp_users.csv';

  logger.info(`Reading copernica profiles from - "${bucket}/${fileName}"`);

  const fileContent = await admin.app().storage().bucket(bucket).file(fileName).download();
  const [headersLine, ...rows] = fileContent.toString().split(LinesSeparator);

  if (dryrun) logger.info('headers', headersLine.split(ColumnsSeparator));

  const copernicaProfiles = rows.map((row) => {
    const columns = row.split(ColumnsSeparator);
    return {
      id: columns[Column.ID],
      uid: columns[Column.UID],
      displayName: columns[Column.DisplayName],
      email: columns[Column.Email],
      isAdmin: columns[Column.IsAdmin],
    };
  });

  let totalUsers = 0;
  let deletedUsers = 0;
  let idIsNull = 0;
  let idIsObject = 0;
  let idIsOk = 0;
  let idsUpdated = 0;
  let missingCopernicaProfiles = 0;

  const handler = async (profileSnap) => {
    const getCopernicaProfile = () => {
      const profiles = copernicaProfiles.filter((p) => p.uid === profileSnap.id);

      if (profiles.length === 0) {
        return null;
      }

      if (profiles.length === 1) {
        return profiles[0];
      }

      // when multiple profiles found we need to pick a correct one.
      if (dryrun) {
        logger.info(`[DRYRUN] [${profileSnap.id}] multiple profiles found`, profiles);
      } else {
        logger.info(`[${profileSnap.id}] multiple profiles found`);
      }

      const profile = profiles.find((p) => p.email || [true, false].includes(p.isAdmin));

      if (dryrun) {
        logger.info(`[DRYRUN] [${profileSnap.id}] picked profile`, profile);
      }

      return profile || profiles[0];
    };

    const updateCopernicaId = async () => {
      logger.info(
        `${dryrun ? '[DRYRUN] ' : ''} [${profileSnap.id}] (${['displayName', '_copernica_id']
          .map((prop) => `${prop}: ${profileSnap.get(prop)}`)
          .join(', ')})`
      );

      const copernicaProfile = getCopernicaProfile();

      if (!copernicaProfile) {
        missingCopernicaProfiles += 1;
        if (dryrun) logger.warn(`[DRYRUN] There is no copernica profile for ${profileSnap.id}`);
        return;
      }

      if (!dryrun) {
        await profileSnap.ref.update({ _copernica_id: copernicaProfile.id });
      }

      idsUpdated += 1;
    };

    totalUsers += 1;

    if (profileSnap.get(DELETED_AT_PROP)) {
      deletedUsers += 1;
      return; // The user is deleted - no need to set copernica id
    }

    const copernicaId = profileSnap.get('_copernica_id');

    if (typeof copernicaId === 'string' && Number(copernicaId) > 0) {
      idIsOk += 1;
      return;
    }

    if (typeof copernicaId === 'undefined' || copernicaId === null) {
      idIsNull += 1;
      await updateCopernicaId();
    } else if (typeof copernicaId === 'object') {
      idIsObject += 1;
      await updateCopernicaId();
    }
  };

  return iterateCollection('profiles', handler, {
    pageSize: 50,
    concurrency: 1,
  }).then(() =>
    logger.info('stats', {
      totalUsers,
      totalCopernicaProfiles: copernicaProfiles.length,
      missingCopernicaProfiles,
      deletedUsers,
      idIsNull,
      idIsObject,
      idIsOk,
      idsUpdated,
    })
  );
};
