const { sha1 } = require('@hatsa/utils/crypto');
const { DocChanges } = require('../../index');
const { fixDate, removeFieldIfNotType, fixPositiveInteger, removeIfExists, deleteField } = require('../fixers');

const { fixProduct } = require('../subtypes/product');
const { fixAssetRefs } = require('../subtypes/assets');

const fixUrlrefRef = (doc) => {
  const urlrefid = doc.get('urlref_id');

  if (urlrefid === 'someUrlRefId') {
    const urlrefUrl = doc.get('urlref_url');
    if (urlrefUrl) return doc.ref.firestore.doc(`urlrefs/${sha1(urlrefUrl)}`);
    return deleteField;
  }

  const urlrefref = doc.get('urlref_ref');

  if (!urlrefid || urlrefref) return undefined;

  return doc.ref.firestore.doc(`urlrefs/${urlrefid}`);
};

const fixUrlrefId = (d) => {
  const urlrefId = d.get('urlref_id');
  const urlrefUrl = d.get('urlref_url');

  if (urlrefId === 'someUrlRefId') {
    if (urlrefUrl) return `urlrefs/${sha1(urlrefUrl)}`;
    return deleteField;
  }
  return undefined;
};

const fixOfferer = (offerer) => {
  if (!offerer) return undefined;

  if (typeof offerer !== 'object') return deleteField;

  // eslint-disable-next-line no-param-reassign
  if (offerer.logo && offerer.logo.startsWith('data:image/')) delete offerer.logo;

  return offerer;
};

const fixUrlref = (urlref) => {
  if (!urlref) return undefined;

  // eslint-disable-next-line no-param-reassign
  if (urlref.offerer && urlref.offerer.logo && urlref.offerer.logo.startsWith('data:image/'))
    delete urlref.offerer.logo;

  return urlref;
};

/**
 * In Previous versions you could switch a product/link to image. This has resulted in some data to be in mode `image'
 * while not having assets.
 *
 * @param item
 * @returns {string} Either link, product or image
 */
const fixMode = (item) => {
  const assets = item.get('assets');
  const mode = item.get('mode');
  if (mode === 'image' && !assets) return 'link';
  return mode || 'product';
};

const fixItem = (item, board, dryrun) => {
  const path = `boards/${item.ref.parent.parent.id}/items/${item.id}`;
  let deleteReason;

  const hasProductId = item.get('urlref_id') || item.get('affiliate_product_id');
  // const deletedAt = item.get('deleted_at');

  if (item.get('mode') === 'link' && !item.get('url') && !hasProductId) {
    deleteReason = `${path} deleting due missing url in link-mode and no linked products`;
  }

  if (item.get('mode') === 'product' && !hasProductId) {
    deleteReason = `${path} deleting due missing 'urlref_id' or 'affiliate_product_id' in product-mode`;
  }

  if (deleteReason) {
    return !dryrun
      ? item.ref.delete().then(() => console.log(deleteReason)) //eslint-disable-line
      : console.log(`DRYRUN: ${deleteReason}`); //eslint-disable-line
  }

  if (item.get('urlref_id') && !item.get('urlref')) {
    // Poor mans 'touch'
    const updatedAt = item.get('updated_at').toDate();
    updatedAt.setSeconds(updatedAt.getSeconds() + 1);
    const touchMsg = `${path} (poor man) "touch" to populate urlref data`;
    return !dryrun
      ? item.ref.update({ updated_at: updatedAt }).then(() => console.log(touchMsg)) //eslint-disable-line
      : console.log(`DRYRUN: ${touchMsg}`); //eslint-disable-line
  }

  return new DocChanges({
    created_at: fixDate(item.get('created_at')),
    created_by: item.get('created_by') || (board && board.get('created_by')),
    // updated_at: fixDate(item.get('updated_at')),
    mode: fixMode(item),
    description: removeFieldIfNotType(item.get('description'), 'string'),
    price: removeFieldIfNotType(item.get('price'), 'number'),
    amount: typeof item.get('amount') !== 'undefined' ? fixPositiveInteger(item.get('amount'), 1) : undefined,

    product: fixProduct(item.get('product')),

    preview: removeIfExists(item.get('preview')), // is only temporary placeholder image
    assets: fixAssetRefs(item),
    offerer: fixOfferer(item.get('offerer')),
    urlref: fixUrlref(item.get('urlref')),

    urlref_ref: fixUrlrefRef(item),
    urlref_id: fixUrlrefId(item),

    // urlref_url  : typeof d.get('urlref_url') === 'string' ? d.get('urlref_url').replace('http://127.0.0.1:2004/?url=', '') : undefined,

    id: removeIfExists(item.get('id')),
    urlref_url: removeIfExists(item.get('urlref_url')),
    __token: removeIfExists(item.get('__token')),
    undefined: removeIfExists(item.get('undefined')),
    _num_likes: typeof item.get('_num_likes') !== 'undefined' ? fixPositiveInteger(item.get('_num_likes')) : undefined,
  });
};

module.exports = fixItem;
