const P = require('bluebird');
const pickBy = require('lodash/pickBy');
const isEqual = require('lodash/isEqual');
const extend = require('lodash/extend');

const { findDocType } = require('../../match');
const { renderChanges } = require('./helpers');

class DocChanges {
  constructor(props) {
    extend(this, props);
  }
}

const getFixerByDocPath = (docPath) => {
  const type = findDocType(docPath);

  if (!type) throw new Error(`no type match for docPath: ${docPath}`);

  let fixer;
  try {
    fixer = require(`./docFixers/doctypes/${type}`);
  } catch (e) {}

  if (typeof fixer !== 'function') throw new Error(`no fixer function found for type: ${type}`);

  return fixer;
};

const fixDoc = (doc, fixer, dryrun, parentDoc) => {
  return P.resolve(fixer(doc, parentDoc, dryrun))
    .then((updates) => {
      if (!(updates instanceof DocChanges)) return {};

      return pickBy(updates, (v, k) => typeof v !== 'undefined' && !isEqual(v, doc.get(k)));
    })
    .tap((updates) => {
      if (dryrun !== false) return;

      if (!Object.keys(updates).length) return;

      return doc.ref.update(updates);
    })
    .then((updates) => ({
      updates,
      dryrun,
      path: doc.ref.path,
      changes: renderChanges(doc, updates),
    }));
};

const fixDocByPath = (firestore, docPath, dryrun) => {
  const fixer = getFixerByDocPath(docPath);
  return firestore
    .doc(docPath)
    .get()
    .then((doc) => fixDoc(doc, fixer, dryrun));
};

const fixDocByRef = (docRef, dryrun) => {
  const fixer = getFixerByDocPath(docRef.path);
  return docRef.get().then((doc) => fixDoc(doc, fixer, dryrun));
};

const fixDocByDoc = (doc, dryrun) => {
  const fixer = getFixerByDocPath(doc.ref.path);
  return fixDoc(doc, fixer, dryrun);
};

const fixDocByDocAndParent = (doc, parentDoc, dryrun) => {
  const fixer = getFixerByDocPath(doc.ref.path);
  return fixDoc(doc, fixer, dryrun, parentDoc);
};

// const fixCollectionByRef  = (colRef) => {};
// const fixCollectionByPath = (firestore, colPath) => {};

module.exports = {
  fixDocByPath,
  fixDocByRef,
  fixDocByDoc,
  fixDocByDocAndParent,

  // fixCollectionByRef,
  // fixCollectionByPath,

  DocChanges,
};
