const P = require('bluebird');
const { CREATED_BY_PROP } = require('@hatsa/constants');
const {
  ROLE_EDITOR_VALUE,
  ROLE_REVOKED_VALUE,
  ROLE_LEFT_VALUE,
  ROLE_VIEWER_VALUE,
  ROLE_OWNER_VALUE,
  getCodeByRoleName,
  ROLE_INVITED_VALUE,
} = require('@hatsa/constants/sharing');
const get = require('lodash/get');
const CollectionPaths = require('../../firestore-paths/CollectionPaths');
const DocumentPaths = require('../../firestore-paths/DocumentPaths');
const { setBoardFolder, changeParticipantRole } = require('../../models/boards');

const { genNotificationFuncs } = require('../notifications');
const { genBoardActivityFuncs } = require('../board-activities');
const { assignDeletedProps, genBoardSlug } = require('../utils');
const { FolderTypes } = require('./folder-types');

const CollectionName = CollectionPaths.Boards();
const ParticipantRoles = [
  ROLE_EDITOR_VALUE,
  ROLE_VIEWER_VALUE,
  ROLE_REVOKED_VALUE,
  ROLE_LEFT_VALUE,
  ROLE_INVITED_VALUE,
];

const ActiveParticipantRoles = [ROLE_EDITOR_VALUE, ROLE_VIEWER_VALUE, ROLE_OWNER_VALUE, ROLE_INVITED_VALUE];

/**
 * Generate common board functions.
 *
 * @param {object} opts
 * @param {FirebaseFirestore.Firestore} opts.firestore
 *
 * @returns {*}
 */
const genBoardFuncs = ({ firestore }) => {
  const { addBoardParticipantRoleChangedNotification } = genNotificationFuncs({ firestore });
  const { addBoardLeftActivity } = genBoardActivityFuncs({ firestore });

  /**
   * Get board reference.
   *
   * @param {string} boardId
   * @returns {FirebaseFirestore.DocumentReference}
   */
  const getBoardRef = (boardId) => firestore.doc(DocumentPaths.Boards({ boardId }));

  /**
   * Generate board item path based on board id or slug and item id.
   *
   * @param {string} boardIdOrSlug
   * @param {string} itemId
   *
   * @returns {Promise<string>}
   */
  const genBoardItemPath = async ({ boardIdOrSlug, itemId }) => {
    if (boardIdOrSlug.includes('-')) {
      // if a string includes "-" - it means it's a board slug
      const boardQuerySnap = await firestore.collection('boards').where('_slug', '==', boardIdOrSlug).limit(1).get();

      if (boardQuerySnap.empty) {
        throw new Error('The board is not found');
      }

      return `boards/${boardQuerySnap.docs[0].id}/items/${itemId}`;
    }

    // otherwise it's a board id
    return `boards/${boardIdOrSlug}/items/${itemId}`;
  };

  /**
   * Check if user has viewer permissions.
   *
   * @param {Object} opts
   * @param {Object} opts.board
   * @param {string} opts.uid
   *
   * @returns boolean
   */
  const isBoardViewer = ({ board, uid }) => {
    const role = get(board, `participants.${uid}.role`, null);
    return role && role >= ROLE_VIEWER_VALUE && role < ROLE_EDITOR_VALUE;
  };

  /**
   * Check if user has editor permissions.
   *
   * @param {Object} opts
   * @param {Object} opts.board
   * @param {string} opts.uid
   *
   * @returns boolean
   */
  const isBoardEditor = ({ board, uid }) => {
    const role = get(board, `participants.${uid}.role`, null);
    return role && role >= ROLE_EDITOR_VALUE && role < ROLE_OWNER_VALUE;
  };

  /**
   * Check if user is the owner of the board.
   *
   * @param {Object} opts
   * @param {Object} opts.board
   * @param {string} opts.uid
   *
   * @returns boolean
   */
  const isBoardOwner = ({ board, uid }) => {
    const createdBy = board[CREATED_BY_PROP];
    return uid && createdBy === uid;
  };

  /**
   * Check if user has at least invitee permissions.
   *
   * @param {Object} opts
   * @param {Object} opts.board
   * @param {string} opts.uid
   *
   * @returns boolean
   */
  const canView = ({ board, uid }) => {
    const isPublic = board && board.public;

    if (isPublic) {
      return true;
    }

    const role = get(board, `participants.${uid}.role`, null);
    return (role && role >= ROLE_INVITED_VALUE) || isBoardOwner({ board, uid });
  };

  /**
   * Check if invite was revoked before user accept/reject it
   * @returns boolean
   */
  const isInvitationRevoked = (oldRoleValue, roleValue) =>
    oldRoleValue === ROLE_INVITED_VALUE && roleValue === ROLE_REVOKED_VALUE;

  /**
   *
   * Fetch board.
   *
   * @param {string|FirebaseFirestore.DocumentReference} boardIdOrRef
   * @returns {Promise<FirebaseFirestore.DocumentSnapshot>}
   */
  const fetchBoard = async (boardIdOrRef) =>
    typeof boardIdOrRef === 'string' ? getBoardRef(boardIdOrRef).get() : boardIdOrRef.get();

  /**
   * Fetch board. Throws an exception when board is not found.
   *
   * @param boardIdOrRef
   * @returns {Promise<FirebaseFirestore.DocumentSnapshot>}
   */
  const fetchBoardOrFail = async (boardIdOrRef) => {
    if (!boardIdOrRef) {
      throw new Error('Board ID or Board reference is required');
    }

    const board = await fetchBoard(boardIdOrRef);

    if (!board.exists) {
      throw new Error('Board does not exist');
    }

    return board;
  };

  /**
   * Fetch user profile.
   *
   * @param {string} userId
   * @returns {Promise<FirebaseFirestore.DocumentSnapshot>}
   */
  const fetchUserProfile = async (userId) => firestore.doc(DocumentPaths.Profiles({ uid: userId })).get();

  /**
   * Create a new board.
   *
   * @param {*} board
   * @returns {Promise<FirebaseFirestore.DocumentReference>}
   */
  const addBoard = async (board) => {
    const newBoardData = board;
    const newBoardRef = firestore.collection(CollectionName).doc();
    newBoardData._slug = genBoardSlug(newBoardData.title, newBoardRef.id);
    await newBoardRef.set(newBoardData);
    return { id: newBoardRef.id, ...newBoardData };
  };

  /**
   * Update board properties.
   *
   * @param {object} opts
   * @param {FirebaseFirestore.DocumentSnapshot} opts.boardSnap
   * @param {string} opts.uid
   * @param {string} opts.folder
   * @param {FirebaseFirestore.Transaction} opts.transaction
   *
   * @returns {Promise<void>}
   */
  const updateBoard = async ({ boardSnap, uid, folder, transaction }) => {
    await transaction.update(boardSnap.ref, setBoardFolder({ boardSnap, uid, folder }));
  };

  /**
   * Update board items.
   *
   * @param {object} opts
   * @param {FirebaseFirestore.DocumentSnapshot} opts.boardSnap
   * @param {string} opts.uid
   * @param {string} opts.folder
   * @param {FirebaseFirestore.Transaction} opts.transaction
   *
   * @returns {Promise<*>}
   */
  const updateBoardItems = async ({ boardSnap, uid, folder, transaction }) => {
    const boardItemsRef = boardSnap.ref.collection('items');
    const boardItemsSnap = await boardItemsRef.get();
    const data =
      folder === FolderTypes.DELETED
        ? assignDeletedProps({ uid, data: { rescrape_until: null } })
        : { rescrape_until: null };

    return boardItemsSnap.docs.map((boardItemSnap) =>
      transaction.set(boardItemsRef.doc(boardItemSnap.id), data, { merge: true })
    );
  };

  /**
   * Add a new item to the board group
   * @param boardData
   * @param groupSlug
   * @param newItemId
   * @return {object[]}
   */
  const updateGroups = ({ boardData, groupSlug, newItemId }) =>
    (boardData.groups || []).map((group) => {
      if (group.slug === groupSlug) {
        return { ...group, items: [...group.items, newItemId] };
      }
      return group;
    });

  /**
   * Notify participants about role changes.
   *
   * @param {object} opts
   * @param {FirebaseFirestore.DocumentSnapshot} opts.boardSnap
   * @param {string} opts.currentUserId
   * @param {string} opts.currentUserDisplayName
   * @param {number} opts.roleValue
   *
   * @returns {Promise<void>}
   */
  const notifyParticipants = async ({ boardSnap, currentUserId, currentUserDisplayName, roleValue }) => {
    const participants = boardSnap.get('participants');
    const participantIds = Object.keys(participants || {}).filter((participantId) => participantId !== currentUserId); // We don't want to notify the user who removed the board

    await Promise.all(
      participantIds.map(async (participantId) => {
        await addBoardParticipantRoleChangedNotification({
          targetUserId: participantId,
          initiatorId: currentUserId,
          initiatorName: currentUserDisplayName || '',
          boardSnap,
          roleValue,
        });
      })
    );
  };

  const isHighlighted = async ({ boardSnap }) => {
    const { id } = boardSnap;
    const slug = boardSnap.get('_slug');
    if (!slug) {
      return false;
    }
    const querySnap = await firestore.collection('offerers').where(`highlighted_boards.${id}._slug`, '==', slug).get();

    return querySnap.size > 0;
  };

  /**
   * Delete board.
   *
   * @param {object} opts
   * @param {string} opts.boardId
   * @param {string} opts.uid
   *
   * @returns {Promise<void>}
   */
  const deleteBoard = async ({ boardId, uid }) =>
    firestore.runTransaction(async (transaction) => {
      const profileSnap = await fetchUserProfile(uid);
      const boardRef = getBoardRef(boardId);
      const boardSnap = await transaction.get(boardRef);
      const args = { boardSnap, uid, folder: FolderTypes.DELETED, transaction };

      await updateBoard(args);
      await updateBoardItems(args);
      await notifyParticipants({
        boardSnap,
        currentUserId: uid,
        currentUserDisplayName: profileSnap.get('displayName') || '',
        roleValue: ROLE_REVOKED_VALUE,
      });
      return { id: boardId };
    });

  /**
   * Archive board.
   *
   * @param {object} opts
   * @param {string} opts.boardId
   * @param {string} opts.uid
   *
   * @returns {Promise<void>}
   */
  const archiveBoard = async ({ boardId, uid }) =>
    firestore.runTransaction(async (transaction) => {
      const profileSnap = await fetchUserProfile(uid);
      const boardRef = getBoardRef(boardId);
      const boardSnap = await transaction.get(boardRef);
      const args = { boardSnap, uid, folder: FolderTypes.ARCHIVED, transaction };

      await updateBoard(args);
      await updateBoardItems(args);
      await notifyParticipants({
        boardSnap,
        currentUserId: uid,
        currentUserDisplayName: profileSnap.get('displayName') || '',
        roleValue: ROLE_REVOKED_VALUE,
      });
    });

  /**
   * Delete user boards.
   *
   * @param {string} userId
   * @param {boolean} [dryrun]
   *
   * @returns {Promise<*[]>}
   */
  const deleteUserBoards = async (userId, dryrun) => {
    const boardsSnap = await firestore.collection(CollectionName).where(CREATED_BY_PROP, '==', userId).get();

    if (boardsSnap.empty) {
      return [];
    }

    const docs = boardsSnap.docs.filter((doc) => !doc.data().deleted_at);

    if (dryrun) {
      return docs.map((doc) => ({ id: doc.id, ...doc.data() }));
    }

    return P.map(
      docs,
      (board) =>
        deleteBoard({
          boardId: board.id,
          uid: userId,
        }),
      { concurrency: 1 }
    );
  };

  /**
   * Get boards where user is participating in.
   *
   * @param {string} userId
   * @returns {Promise<FirebaseFirestore.QuerySnapshot>}
   */
  const getUserParticipatingBoards = async (userId) =>
    firestore.collection(CollectionName).where(`participants.${userId}.role`, '>=', ROLE_VIEWER_VALUE).get();

  /**
   * Update participant role.
   *
   * @param {object} opts
   * @param {FirebaseFirestore.DocumentSnapshot} opts.boardSnap
   * @param {string} opts.roleNameOrValue
   * @param {string} opts.currentUserId
   * @param {string} opts.currentUserDisplayName
   * @param {string} opts.participantUid
   *
   * @returns {Promise<void>}
   */
  const updateParticipantRole = async ({
    boardSnap,
    roleNameOrValue,
    currentUserId,
    currentUserDisplayName,
    participantUid,
  }) => {
    if (isBoardOwner({ board: boardSnap.data(), uid: participantUid })) {
      throw new Error("You are the owner of the board, you can't delete yourself or change the role");
    }

    const roleValue = getCodeByRoleName(roleNameOrValue) || Number(roleNameOrValue);

    if (!roleValue || !ParticipantRoles.includes(roleValue)) {
      throw new Error(`invalid role ${roleNameOrValue}: ${roleValue}`);
    }

    await boardSnap.ref.update(changeParticipantRole({ participantUid, roleValue }));

    if (roleValue === ROLE_LEFT_VALUE) {
      const profileSnap = await fetchUserProfile(participantUid);
      await addBoardLeftActivity({
        boardSnap,
        createdBy: currentUserId,
        participantUid,
        participantName: profileSnap.get('displayName'),
      });
    }
    const oldRoleValue = get(boardSnap.data(), `participants.${participantUid}.role`, null);

    if (isInvitationRevoked(oldRoleValue, roleValue) || currentUserId === participantUid) {
      return;
    }

    await addBoardParticipantRoleChangedNotification({
      targetUserId: participantUid,
      initiatorId: currentUserId,
      initiatorName: currentUserDisplayName || '',
      boardSnap,
      roleValue,
    });
  };

  return {
    getBoardRef,
    addBoard,
    fetchBoard,
    fetchBoardOrFail,
    deleteBoard,
    archiveBoard,
    deleteUserBoards,
    updateParticipantRole,
    getUserParticipatingBoards,
    isBoardViewer,
    isBoardEditor,
    isBoardOwner,
    canView,
    updateGroups,
    isHighlighted,
    genBoardItemPath,
  };
};

module.exports = { genBoardFuncs, CollectionName, ParticipantRoles, ActiveParticipantRoles };
