const { InviteStatuses } = require('@hatsa/constants/invite-statuses');
const { ROLE_REJECTED_VALUE, ROLE_VIEWER_VALUE } = require('@hatsa/constants/sharing');
const get = require('lodash/get');
const { genBoardFuncs } = require('../boards');
const { changeParticipantRole } = require('../../models/boards');
const { genBoardActivityFuncs } = require('../board-activities');
const DocumentPaths = require('../../firestore-paths/DocumentPaths');
const CollectionPaths = require('../../firestore-paths/CollectionPaths');
const { followFuncs } = require('../users/follow-funcs');

/**
 * @typedef {import('firebase-admin').auth.Auth} Auth
 * @typedef {import('firebase-admin').auth.UserRecord} UserRecord
 * @typedef {import('@hatsa/log').HatsaLogger} HatsaLogger
 * @typedef {import('../../schemas/firestore/objects/invites/invite').Invite} Invite
 */

/**
 * Generate invites common functions.
 *
 * @param {object} opts
 * @param {FirebaseFirestore.Firestore} opts.firestore
 * @param {HatsaLogger} [opts.logger]
 */
const genInvitesFuncs = ({ firestore, logger }) => {
  const { fetchBoardOrFail } = genBoardFuncs({ firestore });
  const { addBordJoinedActivity } = genBoardActivityFuncs({ firestore });
  const { followUser } = followFuncs({ firestore });

  /**
   * Get inviter`s invite reference.
   *
   * @param {object} opts
   * @param {string} opts.inviterId
   * @param {string} opts.inviteId
   *
   * @returns {FirebaseFirestore.DocumentReference}
   */
  const getInviteRef = ({ inviterId, inviteId }) =>
    firestore.doc(DocumentPaths.ProfileInvites({ uid: inviterId, inviteId }));

  /**
   * Get profile`s invites collection reference.
   *
   * @param {object} opts
   * @param {string} opts.uid
   *
   * @returns {FirebaseFirestore.CollectionReference}
   */
  const getProfileInvitesRef = ({ uid }) => firestore.collection(CollectionPaths.ProfileInvites({ uid }));

  /**
   * Get profile`s actions collection reference.
   *
   * @param {object} opts
   * @param {string} opts.uid
   *
   * @returns {FirebaseFirestore.CollectionReference}
   */
  const getProfileActionsRef = ({ uid }) => firestore.collection(CollectionPaths.ProfileActions({ uid }));

  /**
   * Fetch inviter`s invite document snapshot.
   *
   * @param {object} opts
   * @param {string|FirebaseFirestore.DocumentReference} opts.inviteIdOrRef
   * @param {string} [opts.inviterId]
   *
   * @returns {Promise<FirebaseFirestore.DocumentSnapshot>}
   */
  const fetchInvite = async ({ inviteIdOrRef, inviterId }) =>
    typeof inviteIdOrRef === 'string'
      ? getInviteRef({ inviterId, inviteId: inviteIdOrRef }).get()
      : inviteIdOrRef.get();

  /**
   * Fetch invites from collection group.
   *
   * @param {object} opts
   * @param {string} opts.inviteId
   */
  const fetchInvitesById = async ({ inviteId }) =>
    firestore.collectionGroup('invites').where('inviteId', '==', inviteId).get();

  /**
   * Fetch pending invites of the user.
   *
   * @param {object} opts
   * @param {string} opts.inviterId
   * @param {string} opts.email
   *
   * @returns {Promise<FirebaseFirestore.QuerySnapshot>}
   */
  const fetchPendingInvites = async ({ inviterId, email }) => {
    const invitesRef = getProfileInvitesRef({ uid: inviterId });

    return invitesRef.where('email', '==', email).where('status', '==', InviteStatuses.Sent).get();
  };

  /**
   * Check if invite is valid and in correct status to be accepted or rejected.
   *
   * @param {object} inviteData
   *
   * @returns {boolean}
   */
  const isValidInvite = (inviteData) => {
    const { status } = inviteData || {};

    if (status !== InviteStatuses.Sent) {
      if (logger) logger.warn(`Invite has invalid status - '${status}'. Expected to be - '${InviteStatuses.Sent}'`);
      return false;
    }

    return true;
  };

  /**
   * Check if invite of type board is valid.
   *
   * @param {object} inviteData
   *
   * @returns {boolean}
   */
  const isValidBoardInvite = (inviteData) => {
    if (!inviteData || (!inviteData.boardId && !inviteData.boardid)) {
      if (logger) logger.warn('`boardId` is missing in the invite document', inviteData);
      return false;
    }

    return isValidInvite(inviteData);
  };

  /**
   * Change profile`s actions statuses related to the given invite.
   *
   * @param {object} opts
   * @param {string} opts.uid
   * @param {string} opts.inviteId
   * @param {string} opts.status
   *
   * @returns {Promise<void>}
   */
  const changeActionStatus = async ({ uid, inviteId, status, failureReason = null }) => {
    const actionsSnap = await getProfileActionsRef({ uid }).where('type', '==', 'board_invite').get();

    if (actionsSnap.empty) {
      return;
    }

    const setData = failureReason ? { status, failureReason } : { status };
    const actionDocs = actionsSnap.docs.filter((actionSnap) => actionSnap.get('data.invite.id') === inviteId);

    if (failureReason) {
      logger.warn(`Failed to accept or reject an action`, {
        failureReason,
        actions: actionDocs.map((actionSnap) => ({
          docPath: `/profiles/${uid}/actions/${actionSnap.id}`,
        })),
      });
    }

    await Promise.all(
      actionDocs.map((actionSnap) => actionSnap.ref.set({ data: { invite: setData } }, { merge: true }))
    );
  };

  /**
   * Accept profile`s actions related to the given invite.
   *
   * @param {object} opts
   * @param {string} opts.uid
   * @param {string} opts.inviteId
   *
   * @returns {Promise<void>}
   */
  const acceptActions = async ({ uid, inviteId }) =>
    changeActionStatus({ uid, inviteId, status: InviteStatuses.Accepted });

  /**
   * Reject profile`s actions related to the given invite.
   *
   * @param {object} opts
   * @param {string} opts.uid
   * @param {string} opts.inviteId
   *
   * @returns {Promise<void>}
   */
  const rejectActions = async ({ uid, inviteId }) =>
    changeActionStatus({ uid, inviteId, status: InviteStatuses.Rejected });

  /**
   * Mark actions related to given invite as failed
   *
   * @param {object} opts
   * @param {string} opts.uid
   * @param {string} opts.inviteId
   *
   * @returns {Promise<void>}
   */
  const markFailedActions = async ({ uid, inviteId, failureReason }) =>
    changeActionStatus({ uid, inviteId, status: InviteStatuses.Failed, failureReason });

  /**
   * Add additional info about the invitee user.
   *
   * @param inviteRef
   * @param status
   * @param invitee
   * @param inviteData
   *
   * @returns {Promise<void>}
   */
  const updateInvite = async ({ inviteRef, status, invitee, inviteData }) => {
    const inviteChange = { status };

    if (!inviteData.invitee || typeof inviteData.invitee === 'string') {
      // invitee data can be null if invitee did not have account when invite comes
      inviteChange.invitee = {
        uid: invitee.uid,
        displayName: invitee.displayName || '',
        email: invitee.email || null,
        photoURL: invitee.photoURL || null,
      };
    }

    await inviteRef.set(inviteChange, { merge: true });
  };

  const _checkUserCanAcceptOrRejectInvite = ({ inviteSnap, invitee }) => {
    const inviterId = get(inviteSnap.data(), 'inviter.uid', null);

    if (inviterId && inviterId === invitee.uid) {
      throw new Error('You can not accept/reject your own invites');
    }
  };

  /**
   * Accept board invite.
   *
   * @param {object} opts
   * @param {FirebaseFirestore.DocumentSnapshot} opts.inviteSnap
   * @param {UserRecord} opts.invitee
   *
   * @returns {Promise<*>}
   */
  const acceptBoardInvite = async ({ inviteSnap, invitee }) => {
    const inviteData = inviteSnap.data() || {};
    const boardId = inviteSnap.get('boardid') || inviteSnap.get('boardId');
    const { inviter, role } = inviteData;

    if (!isValidBoardInvite(inviteData)) {
      throw new Error('Can not accept invite - the invite is not valid');
    }

    const boardSnap = await fetchBoardOrFail(boardId);
    await boardSnap.ref.update(
      changeParticipantRole({
        participantUid: invitee.uid,
        roleValue: role || ROLE_VIEWER_VALUE,
      })
    );

    await updateInvite({
      inviteRef: inviteSnap.ref,
      status: InviteStatuses.Accepted,
      invitee,
      inviteData,
    });
    await acceptActions({ uid: invitee.uid, inviteId: inviteSnap.id });
    await addBordJoinedActivity({ boardSnap, invitee, inviter });

    return { id: inviteSnap.id, ...(inviteSnap.data() || {}) };
  };

  /**
   * Accept profile invite.
   *
   * @param {object} opts
   * @param {FirebaseFirestore.DocumentSnapshot} opts.inviteSnap
   * @param {UserRecord} opts.invitee
   *
   * @returns {Promise<*>}
   */
  const acceptProfileInvite = async ({ inviteSnap, invitee }) => {
    const inviteData = inviteSnap.data() || {};
    const { inviter, followBack } = inviteData;
    const { uid: inviteeId } = invitee;
    const { uid: inviterId } = inviter;

    if (!isValidInvite(inviteData)) {
      throw new Error('Can not accept invite - the invite is not valid');
    }

    await updateInvite({
      inviteRef: inviteSnap.ref,
      status: InviteStatuses.Accepted,
      invitee,
      inviteData,
    });
    await acceptActions({ uid: invitee.uid, inviteId: inviteSnap.id });
    await followUser({ followerUid: inviteeId, followingUid: inviterId });

    if (followBack) {
      await followUser({ followerUid: inviterId, followingUid: inviteeId });
    }

    return { id: inviteSnap.id, ...(inviteSnap.data() || {}) };
  };

  /**
   * Accept invite.
   *
   * @param {object} opts
   * @param {FirebaseFirestore.DocumentSnapshot} opts.inviteSnap
   * @param {UserRecord} opts.invitee
   *
   * @returns {Promise<*>}
   */
  const acceptInvite = async ({ inviteSnap, invitee }) => {
    try {
      const type = inviteSnap.get('type');
      _checkUserCanAcceptOrRejectInvite({ inviteSnap, invitee });

      switch (type) {
        case 'board':
          return await acceptBoardInvite({ inviteSnap, invitee });
        case 'profile':
          return await acceptProfileInvite({ inviteSnap, invitee });
        default:
          throw new Error(`Unknown invite type - ${type}`);
      }
    } catch (e) {
      //  Mark inviter's invite and invitee's actions as failed
      await Promise.all(
        inviteSnap.exists
          ? [
              inviteSnap.ref.set({ status: InviteStatuses.Failed }, { merge: true }),
              markFailedActions({
                uid: invitee.uid,
                inviteId: inviteSnap.id,
                failureReason: `Failed to accept an invite - ${e.message}`,
              }),
            ]
          : [
              markFailedActions({
                uid: invitee.uid,
                inviteId: inviteSnap.id,
                failureReason: `Failed to accept an invite - invite is not found ${inviteSnap.id}`,
              }),
            ]
      );
      return null;
    }
  };

  /**
   * Reject invite of board type.
   *
   * @param {object} opts
   * @param {FirebaseFirestore.DocumentSnapshot} opts.inviteSnap
   * @param {UserRecord} opts.invitee
   *
   * @returns {Promise<*>}
   */
  const rejectBoardInvite = async ({ inviteSnap, invitee }) => {
    const inviteData = inviteSnap.data() || {};
    const boardId = inviteSnap.get('boardid') || inviteSnap.get('boardId');

    if (!isValidBoardInvite(inviteData)) {
      throw new Error('Can not reject invite - the invite is not valid');
    }
    const boardSnap = await fetchBoardOrFail(boardId);
    await boardSnap.ref.update(
      changeParticipantRole({
        participantUid: invitee.uid,
        roleValue: ROLE_REJECTED_VALUE,
      })
    );

    await inviteSnap.ref.set({ status: InviteStatuses.Rejected }, { merge: true });
    await rejectActions({ uid: invitee.uid, inviteId: inviteSnap.id });

    return { id: inviteSnap.id, ...(inviteSnap.data() || {}) };
  };

  /**
   * Reject invite of profile type.
   *
   * @param {object} opts
   * @param {FirebaseFirestore.DocumentSnapshot} opts.inviteSnap
   * @param {UserRecord} opts.invitee
   *
   * @returns {Promise<*>}
   */
  const rejectProfileInvite = async ({ inviteSnap, invitee }) => {
    const inviteData = inviteSnap.data() || {};

    if (!isValidInvite(inviteData)) {
      throw new Error('Can not reject invite - the invite is not valid');
    }

    await inviteSnap.ref.set({ status: InviteStatuses.Rejected }, { merge: true });
    await rejectActions({ uid: invitee.uid, inviteId: inviteSnap.id });

    return { id: inviteSnap.id, ...(inviteSnap.data() || {}) };
  };

  /**
   * Reject invite.
   *
   * @param {object} opts
   * @param {FirebaseFirestore.DocumentSnapshot} opts.inviteSnap
   * @param {UserRecord} opts.invitee
   *
   * @returns {Promise<*>}
   */
  const rejectInvite = async ({ inviteSnap, invitee }) => {
    try {
      const type = inviteSnap.get('type');
      _checkUserCanAcceptOrRejectInvite({ inviteSnap, invitee });

      switch (type) {
        case 'board':
          return await rejectBoardInvite({ inviteSnap, invitee });
        case 'profile':
          return await rejectProfileInvite({ inviteSnap, invitee });
        default:
          throw new Error(`Unknown invite type - ${type}`);
      }
    } catch (e) {
      //  Mark inviter's invite and invitee's actions as failed
      await Promise.all(
        inviteSnap.exists
          ? [
              inviteSnap.ref.set({ status: InviteStatuses.Failed }, { merge: true }),
              markFailedActions({
                uid: invitee.uid,
                inviteId: inviteSnap.id,
                failureReason: `Failed to reject an invite - ${e.message}`,
              }),
            ]
          : [
              markFailedActions({
                uid: invitee.uid,
                inviteId: inviteSnap.id,
                failureReason: `Failed to reject an invite - invite is not found ${inviteSnap.id}`,
              }),
            ]
      );

      return null;
    }
  };

  /**
   * Fetch invites from collection group.
   *
   * @param {object} opts
   * @param {string} opts.inviteId
   *
   * @returns {Promise<FirebaseFirestore.QuerySnapshot>}
   */
  const fetchInvitesByIdOrFail = async ({ inviteId }) => {
    const invitesSnap = await fetchInvitesById({ inviteId });

    if (invitesSnap.empty) {
      throw new Error(`No invites found for inviteId - '${inviteId}'`);
    }

    if (invitesSnap.size > 1) {
      if (logger) {
        logger.warn(
          'Multiple invites found',
          invitesSnap.docs.map((inviteSnap) => ({
            id: inviteSnap.get('id'),
            boardId: inviteSnap.get('boardId'),
            email: inviteSnap.get('email'),
          }))
        );
      }
    }

    return invitesSnap;
  };

  /**
   * Accept invitation.
   *
   * @param {object} opts
   * @param {string} opts.inviteId
   * @param {UserRecord} opts.invitee
   * @param {string} [opts.inviterId]
   *
   * @returns {Promise<Array<null | { id: string, boardId: string }>|{ id: string, boardId: string }>}
   */
  const accept = async ({ inviteId, invitee, inviterId }) => {
    if (inviterId) {
      // If inviter id is present - no need to make a collection group request,
      // we can get invite directly from the inviter`s profile.
      return acceptInvite({
        inviteSnap: await fetchInvite({ inviteIdOrRef: inviteId, inviterId }),
        invitee,
      });
    }

    const invitesSnap = await fetchInvitesByIdOrFail({ inviteId });

    // In most cases we'll have only one invite.
    // We're using collection group query, so there's a small chance that we'll get multiple invites.
    // So we need to iterate over all docs to catch corner cases.
    return Promise.all(
      invitesSnap.docs.map((inviteSnap) =>
        acceptInvite({
          inviteSnap,
          invitee,
        })
      )
    );
  };

  /**
   * Reject invitation.
   *
   * @param {object} opts
   * @param {string} opts.inviteId
   * @param {UserRecord} opts.invitee
   * @param {string} [opts.inviterId]
   *
   * @returns {Promise<Array<null | { id: string, boardId: string }>|{ id: string, boardId: string }>}
   */
  const reject = async ({ inviteId, invitee, inviterId }) => {
    if (inviterId) {
      // If inviter id is present - no need to make a collection group request,
      // we can get invite directly from the inviter`s profile.
      return rejectInvite({
        inviteSnap: await fetchInvite({ inviteIdOrRef: inviteId, inviterId }),
        invitee,
      });
    }

    const invitesSnap = await fetchInvitesByIdOrFail({ inviteId });

    // In most cases we'll have only one invite.
    // We're using collection group query, so there's a small chance that we'll get multiple invites.
    // So we need to iterate over all docs to catch corner cases.
    return Promise.all(
      invitesSnap.docs.map((inviteSnap) =>
        rejectInvite({
          inviteSnap,
          invitee,
        })
      )
    );
  };

  return {
    fetchPendingInvites,
    fetchInvitesByIdOrFail,
    accept,
    reject,
  };
};

module.exports = { genInvitesFuncs };
