const { roundTime } = require('@hatsa/utils/date/date-time');
const CollectionPaths = require('../../firestore-paths/CollectionPaths');

const { createRoleChangedNotification, createBoardReminderNotification } = require('../../models/notifications');
const models = require('../../models');
const {
  createEventNotApplied,
  createEventNotDrawn,
  createEventIsOver,
  createEventBoardEmpty,
} = require('../../models/notifications/event');

/**
 * Generate notifications functions.
 *
 * @param {object} opts
 * @param {FirebaseFirestore.Firestore} opts.firestore
 */
const genNotificationFuncs = ({ firestore, logger, dryRun }) => {
  /**
   * Get profile notifications collection reference.
   *
   * @param uid
   * @returns {FirebaseFirestore.CollectionReference}
   */
  const getNotificationsCollectionRef = (uid) => firestore.collection(CollectionPaths.ProfileNotifications({ uid }));

  /**
   * Add notification to the database if it does not exist.
   *
   * @param {object} opts
   * @param {FirebaseFirestore.CollectionReference} opts.notificationsRef
   * @param {string} opts.docId
   * @param {object} opts.notification
   * @param {FirebaseFirestore.WriteBatch} batch
   *
   * @returns {Promise<*>}
   */
  const upsert = async ({ notificationsRef, docId, notification }, { batch } = {}) => {
    const notificationRef = notificationsRef.doc(docId);

    if (dryRun) {
      logger?.debug('notification', { docPath: notificationsRef.path, docId, notification });
      return null;
    }

    if (batch) {
      return batch.set(notificationRef, notification);
    }

    return firestore.runTransaction(async (t) => {
      const notificationSnap = await t.get(notificationRef);

      if (notificationSnap.exists) {
        return null;
      }

      return t.set(notificationRef, notification);
    });
  };

  /**
   * Add notification about board participant role changes.
   *
   * @param {object} opts
   * @param {string} opts.targetUserId - The user to whom the notification is addressed.
   * @param {string} opts.initiatorId - The ID of the user who initiates the notification.
   * @param {string} opts.initiatorName - The display name of the user who initiates the notification.
   * @param {FirebaseFirestore.DocumentSnapshot} opts.boardSnap - The board snapshot.
   * @param {number} opts.roleValue - The value of the changed role.
   *
   * @returns {Promise<*>}
   */
  const addBoardParticipantRoleChangedNotification = async ({
    targetUserId,
    initiatorId,
    initiatorName,
    boardSnap,
    roleValue,
  }) => {
    const notificationsRef = getNotificationsCollectionRef(targetUserId);
    const docId = ['board_role', roleValue, boardSnap.id, initiatorId].join('_');
    const notification = createRoleChangedNotification({
      initiatorId,
      initiatorName,
      boardSnap,
      notificationsRef,
      roleValue,
    });

    return upsert({ notificationsRef, docId, notification });
  };

  /**
   * Add notification to remind board owner about old board.
   */
  const addBoardReminder = async ({ uid, board }, { batch } = {}) => {
    const notificationsRef = getNotificationsCollectionRef(uid);
    const today = new Date().toISOString().split('T')[0].replace(/-/g, '_');
    const docId = ['board_reminder', board.id, today].join('_');
    const notification = createBoardReminderNotification({ board, notificationsRef });

    return upsert({ notificationsRef, docId, notification }, { batch });
  };

  /**
   * Notify event owner when not all participants applied the event.
   */
  const addEventNotApplied = async ({ ownerId, event }, { batch } = {}) => {
    const notificationsRef = getNotificationsCollectionRef(ownerId);
    const today = new Date().toISOString().split('T')[0].replace(/-/g, '_');
    const docId = ['event_not_applied', event.id, today].join('_');
    const notification = createEventNotApplied({ event, notificationsRef });

    return upsert({ notificationsRef, docId, notification }, { batch });
  };

  /**
   * Notify event owner when all participants applied but names are not drawn.
   */
  const addEventNotDrawn = async ({ ownerId, event }, { batch } = {}) => {
    const notificationsRef = getNotificationsCollectionRef(ownerId);
    const today = new Date().toISOString().split('T')[0].replace(/-/g, '_');
    const docId = ['event_not_drawn', event.id, today].join('_');
    const notification = createEventNotDrawn({ event, notificationsRef });

    return upsert({ notificationsRef, docId, notification }, { batch });
  };

  /**
   * Notify the event participant when his board is empty.
   */
  const addEventBoardEmpty = async ({ participantId, event, board }, { batch } = {}) => {
    const notificationsRef = getNotificationsCollectionRef(participantId);
    const today = new Date().toISOString().split('T')[0].replace(/-/g, '_');
    const docId = ['event_board_empty', event.id, today].join('_');
    const notification = createEventBoardEmpty({ notificationsRef, event, board });

    return upsert({ notificationsRef, docId, notification }, { batch });
  };

  /**
   * Notify event participant when event is over.
   */
  const addEventIsOver = async ({ participantId, event, board }, { batch } = {}) => {
    const notificationsRef = getNotificationsCollectionRef(participantId);
    const today = new Date().toISOString().split('T')[0].replace(/-/g, '_');
    const docId = ['event_over', event.id, today].join('_');
    const notification = createEventIsOver({ notificationsRef, event, board });

    return upsert({ notificationsRef, docId, notification }, { batch });
  };

  const addUserStartFollowing = async ({ followingUid, followerUserSnap }) => {
    const notificationsRef = getNotificationsCollectionRef(followingUid);
    const { createFollowNotification } = models(notificationsRef);
    const roundedTime = roundTime(1000 * 60 * 60);
    const docId = ['follow', followerUserSnap.id, roundedTime].join('-');
    const notification = createFollowNotification({
      user: {
        uid: followerUserSnap.id,
        displayName: followerUserSnap.get('displayName'),
        photoURL: followerUserSnap.get('photoURL'),
      },
    });

    return upsert({ notificationsRef, docId, notification });
  };

  return {
    addBoardParticipantRoleChangedNotification,
    addBoardReminder,
    addUserStartFollowing,
    addEventNotApplied,
    addEventNotDrawn,
    addEventIsOver,
    addEventBoardEmpty,
  };
};

module.exports = { genNotificationFuncs };
