const { CREATED_AT_PROP } = require('@hatsa/constants');
const { NUM_FOLLOWERS_PROP, NUM_FOLLOWING_PROP } = require('@hatsa/constants/profile');
const DocumentPaths = require('../../firestore-paths/DocumentPaths');
const getFsSentinel = require('../../sentinel');
const { genNotificationFuncs } = require('../notifications');

/**
 * Generate user`s follow functions.
 *
 * @param {object} opts
 * @param {FirebaseFirestore.Firestore} opts.firestore
 * @param {HatsaLogger} [opts.logger]
 */
const followFuncs = ({ firestore, logger }) => {
  const { addUserStartFollowing } = genNotificationFuncs({ firestore });

  const throwIfNotExist = (snap, message) => {
    if (!snap.exists) {
      throw new Error(message);
    }
  };

  /**
   * Follow/unfollow user.
   *
   * @param followerUid
   * @param followingUid
   * @param follow
   *
   * @returns {Promise<void>}
   */
  const performFollow = async ({ followerUid, followingUid, follow }) =>
    firestore.runTransaction(async (t) => {
      const followerUserRef = firestore.doc(DocumentPaths.Profiles({ uid: followerUid }));
      const followingUserRef = firestore.doc(DocumentPaths.Profiles({ uid: followingUid }));
      const followingDocRef = followerUserRef.collection('following').doc(followingUid);
      const followerDocRef = followingUserRef.collection('followers').doc(followerUid);
      const mutation = follow ? 1 : -1;
      const { increment } = getFsSentinel(firestore);

      const [followerUserSnap, followingUserSnap] = await Promise.all([
        t.get(followerUserRef),
        t.get(followingUserRef),
      ]);

      throwIfNotExist(followerUserSnap, 'The follower profile does not exist');
      throwIfNotExist(followingUserSnap, 'The following profile does not exist');

      const [followerDocSnap, followingDocSnap] = await Promise.all([t.get(followerDocRef), t.get(followingDocRef)]);

      if (follow && (followerDocSnap.exists || followingDocSnap.exists)) {
        if (logger) logger.info('follow data already found', { followerUid, followingUid });
        return;
      }

      if (!follow && (!followerDocSnap.exists || !followingDocSnap.exists)) {
        if (logger) logger.info('follow data not found', { followerUid, followingUid });
        return;
      }

      await Promise.all([
        follow ? t.set(followerDocRef, { [CREATED_AT_PROP]: new Date() }) : t.delete(followerDocRef),
        follow ? t.set(followingDocRef, { [CREATED_AT_PROP]: new Date() }) : t.delete(followingDocRef),
        t.update(followerUserRef, { [NUM_FOLLOWING_PROP]: increment(mutation) }),
        t.update(followingUserRef, { [NUM_FOLLOWERS_PROP]: increment(mutation) }),
      ]);

      if (follow) {
        await addUserStartFollowing({ followingUid, followerUserSnap });
      }
    });

  const followUser = async ({ followerUid, followingUid }) =>
    performFollow({ followerUid, followingUid, follow: true });

  const unFollowUser = async ({ followerUid, followingUid }) =>
    performFollow({ followerUid, followingUid, follow: false });

  return {
    followUser,
    unFollowUser,
  };
};

module.exports = { followFuncs };
