const { ROLE_LEFT_VALUE } = require('@hatsa/constants/sharing');
const getFsSentinel = require('../../sentinel');
const DocumentPaths = require('../../firestore-paths/DocumentPaths');
const { setProfileDeleted } = require('../../models/profiles');
const { createBoardLeftActivity } = require('../../models/boards/board-activities');
const { genBoardFuncs } = require('../boards');
const { genBoardActivityFuncs } = require('../board-activities');
const { changeParticipantRole } = require('../../models/boards');

const genUserFuncs = ({ firestore }) => {
  const { deleteUserBoards, getUserParticipatingBoards } = genBoardFuncs({ firestore });
  const { getActivitiesCollectionRef } = genBoardActivityFuncs({ firestore });

  const getProfileRef = (uid) => firestore.doc(DocumentPaths.Profiles({ uid }));

  /**
   * Fetch board.
   *
   * @param {string|DocumentReference} userIdOrRef
   * @returns {Promise<FirebaseFirestore.DocumentSnapshot>}
   */
  const fetchUser = async (userIdOrRef) =>
    typeof userIdOrRef === 'string' ? getProfileRef(userIdOrRef).get() : userIdOrRef.get();

  /**
   * Strip profile props.
   *
   * @param {FirebaseFirestore.DocumentSnapshot} profileSnap
   *
   * @returns {Promise<*>}
   */
  const stripProfileData = async (profileSnap) => {
    const deletedProfile = setProfileDeleted({ profileSnap });
    return profileSnap.ref.set(deletedProfile);
  };

  /**
   * The user leaves boards he is participating in.
   *
   * @param {FirebaseFirestore.DocumentSnapshot} profileSnap
   * @param {boolean} [dryrun]
   *
   * @returns {Promise<*[]>}
   */
  const leaveParticipatingBoards = async (profileSnap, dryrun) => {
    const batch = firestore.batch();
    const userId = profileSnap.id;
    const boardsSnap = await getUserParticipatingBoards(userId);

    if (boardsSnap.empty) {
      return [];
    }

    const boardsData = boardsSnap.docs.map((doc) => ({ id: doc.id, ...doc.data() }));

    if (dryrun) return boardsData;

    boardsSnap.forEach((boardSnap) => {
      batch.update(
        boardSnap.ref,
        changeParticipantRole({
          participantUid: userId,
          roleValue: ROLE_LEFT_VALUE,
        })
      );

      batch.set(
        getActivitiesCollectionRef(boardSnap.id).doc(),
        createBoardLeftActivity({
          boardSnap,
          createdBy: userId,
          participantUid: userId,
          participantName: profileSnap.get('displayName') || '',
        })
      );
    });

    await batch.commit();
    return boardsData;
  };

  /**
d     * Delete documents with batches.
     *
     * @param query
     * @param addBatchItems
     * @param resolve
     * @returns {Promise<void>}
     */
  const deleteQueryBatch = async ({ query, addBatchItems, resolve }) => {
    const batchSize = 500;
    const snapshot = await query.limit(batchSize).get();
    const docsFound = snapshot.size;
    if (docsFound === 0) {
      resolve();
      return;
    }
    const batch = firestore.batch();
    snapshot.docs.forEach((doc) => {
      addBatchItems(batch, doc);
    });
    await batch.commit();

    process.nextTick(() => {
      deleteQueryBatch({ query, addBatchItems, resolve });
    });
  };

  /**
   * Delete user records from other users followers / following sub collections
   * Update profiles counters for followers / following
   * @param userId
   * @returns {Promise<Promise<unknown>[]>}
   */
  const deleteUserFollowRelations = async (userId) => {
    const { increment } = getFsSentinel(firestore);
    const profileDocRef = firestore.collection('profiles').doc(userId);
    const followingQuery = profileDocRef.collection('following').orderBy('created_at');
    const followersQuery = profileDocRef.collection('followers').orderBy('created_at');
    return Promise.all([
      new Promise((resolve, reject) => {
        const addBatchItems = (batch, doc) => {
          const { id: followingUserId } = doc;
          batch.delete(firestore.collection('profiles').doc(followingUserId).collection('followers').doc(userId));
          batch.delete(doc.ref);
          batch.update(firestore.collection('profiles').doc(followingUserId), {
            _num_followers: increment(-1),
          });
        };
        deleteQueryBatch({ firestore, query: followingQuery, addBatchItems, resolve }).catch(reject);
      }).catch(() => {}),
      new Promise((resolve, reject) => {
        const addBatchItems = (batch, doc) => {
          const { id: followerUserId } = doc;
          batch.delete(firestore.collection('profiles').doc(followerUserId).collection('following').doc(userId));
          batch.delete(doc.ref);
          batch.update(firestore.collection('profiles').doc(followerUserId), {
            _num_following: increment(-1),
          });
        };
        deleteQueryBatch({ firestore, query: followersQuery, addBatchItems, resolve }).catch(reject);
      }).catch(() => {}),
    ]);
  };

  /**
   * Delete different active notifications of deleted user
   *
   * @param userId
   * @return {Promise<Promise<unknown>[]>}
   */
  const deleteNotifications = async (userId) => {
    const actionsQuery = await firestore
      .collectionGroup('actions')
      .where('data.inviter.uid', '==', userId)
      .where('data.invite.status', '==', 'sent');

    const deleteAllActiveInvitations = new Promise((resolve, reject) => {
      const addBatchItems = (batch, doc) => {
        batch.delete(doc.ref);
      };
      deleteQueryBatch({ firestore, query: actionsQuery, addBatchItems, resolve }).catch(reject);
    });
    await deleteAllActiveInvitations;
    return null;
  };

  /**
   * Delete user profile and related data.
   *
   * @param {string} userId
   * @returns {Promise<*>}
   */
  const deleteUser = async ({ userId }) => {
    const profileSnap = await getProfileRef(userId).get();

    await Promise.all([
      stripProfileData(profileSnap),
      deleteUserBoards(userId),
      leaveParticipatingBoards(profileSnap),
      deleteUserFollowRelations(userId),
      deleteNotifications(userId),
    ]);
    return null;
  };

  return {
    fetchUser,
    deleteUser,
    leaveParticipatingBoards,
  };
};

module.exports = { genUserFuncs };
