const {
  DELETED_AT_PROP,
  DELETED_BY_PROP,
  UPDATED_AT_PROP,
  UPDATED_BY_PROP,
  CREATED_AT_PROP,
  CREATED_BY_PROP,
} = require('@hatsa/constants');
const { isDifferentOnly } = require('../isDifferent');
const getSentinels = require('../sentinel');
const slugify = require('@hatsa/utils/url/slugify');

/**
 * Assign `created_at` and `created_by` props.
 *
 * @param {object} opts
 * @param {string} [opts.uid]
 * @param {object} opts.data
 *
 * @returns {{object}}
 */
const assignCreatedProps = ({ uid, data = {} }) => ({
  ...data,
  [CREATED_AT_PROP]: new Date(),
  ...(uid ? { [CREATED_BY_PROP]: uid } : undefined),
});

/**
 * Assign `deleted_at` and `deleted_by` props.
 *
 * @param {object} opts
 * @param {string} [opts.uid]
 * @param {object} opts.data
 *
 * @returns {{[p: string]: Date, deleted_by?: string, '[DELETED_AT_PROP]': Date}}
 */
const assignDeletedProps = ({ uid, data = {} }) => ({
  ...data,
  [DELETED_AT_PROP]: new Date(),
  ...(uid ? { [DELETED_BY_PROP]: uid } : undefined),
});

/**
 * Assign `updated_at` and `updated_by` props.
 *
 * @param {object} opts
 * @param {string} [opts.uid]
 * @param {object} opts.data
 *
 * @returns {{[p: string]: Date, '[UPDATED_AT_PROP]': Date, updated_by?: string}}
 */
const assignUpdatedProps = ({ uid, data = {} }) => ({
  ...data,
  [UPDATED_AT_PROP]: new Date(),
  ...(uid ? { [UPDATED_BY_PROP]: uid } : undefined),
});

/**
 * Checks if document was affected by migration
 * @param change
 * @return {boolean}
 */
const isUpdatedByMigration = (change) => {
  if (change.after.get('_migration') || isDifferentOnly(change, '_migration')) {
    return true;
  }
  return false;
};

/**
 * Remove migration state field.
 * @param change
 * @return {Promise<boolean>}
 */
const removeMigrationStateField = async (change) => {
  const { deleteField } = getSentinels(change.after.ref.firestore);
  if (!change.after.get('_migration')) {
    return false;
  }
  await change.after.ref.update({ _migration: deleteField });
  return true;
};

const genBoardSlug = (title, id) => {
  if (typeof title !== 'string' || typeof id !== 'string') {
    return '';
  }
  return `${slugify(title)}-${id.substring(0, 6)}`;
};

module.exports = {
  assignDeletedProps,
  assignUpdatedProps,
  assignCreatedProps,
  isUpdatedByMigration,
  removeMigrationStateField,
  genBoardSlug,
};
