const P = require('bluebird');
const whilst = require('async/whilst');
const castArray = require('lodash/castArray');
const { Readable } = require('stream');

const genIteratorFuncs = (firestore) => {
  const documentId = firestore.constructor.FieldPath.documentId();

  const iterate =
    (cType, sType) =>
    (col, itemCb, opts = {}) => {
      const asPromised = !sType || sType === 'promise';
      const readableStream = !asPromised && new Readable({ objectMode: true, read() {} });

      const {
        orderBy = documentId,
        orderDir = 'asc',
        pageSize = 10,
        concurrency = 10,
        select,
        where,
        query: givenQuery,
        logger,
      } = opts;

      let query;

      if (!givenQuery) {
        query = firestore[cType](col).orderBy(orderBy, orderDir).limit(pageSize);

        if (select) query = query.select(...castArray(select));

        if (Array.isArray(where)) {
          const isMultiWhere = where.length && Array.isArray(where[0]);

          if (isMultiWhere)
            where.forEach((w) => {
              query = query.where(...w);
            });
          else query = query.where(...where);
        }
      } else query = givenQuery;

      const { collectionId: _col, limit: _limit } = query._queryOptions;

      let length = null;
      let lastDoc = null;
      const returnDocs = [];

      function start(resolve, reject) {
        whilst(
          // test func:
          () => length > 0 || typeof length !== 'number',
          // iteratee func:
          (callback) => {
            const queryToFetch = lastDoc ? query.startAfter(lastDoc) : query;

            if (logger)
              logger.info(
                `fetching ${_limit} firestore docs from ${_col} ${lastDoc ? `(start after ${lastDoc.id})` : ''}`
              );

            if (!asPromised) {
              length = 0;
              return queryToFetch
                .stream()
                .on('data', (data) => {
                  lastDoc = data;
                  length++;
                  readableStream.push(data);
                })
                .on('end', callback)
                .on('error', callback);
            }

            // console.log(`new page with ${pageSize} items`);
            return P.resolve(queryToFetch.get())
              .then((ss) => ss.docs)
              .tap((docs) => {
                length = docs.length;
                lastDoc = docs[length - 1];

                if (itemCb) {
                  returnDocs.push(...docs.map((d) => d.id));
                  return P.map(docs, itemCb, { concurrency });
                }

                return returnDocs.push(...docs);
              })
              .nodeify(callback);
          },
          // finish callback func:
          (err) => {
            if (asPromised) return err ? reject(err) : resolve(returnDocs);
            if (err) return readableStream.emit('error', err);

            return readableStream.push(null);
          }
        );
      }

      if (asPromised) return new P(start);

      start();
      return readableStream;
    };

  const iterateCollection = iterate('collection');
  const iterateCollectionGroup = iterate('collectionGroup');

  const streamCollection = iterate('collection', 'stream');
  const streamCollectionGroup = iterate('collectionGroup', 'stream');

  const iterateSubCollection = (c1, c2, handler, c1opts, c2opts) => {
    const result = {};
    // eslint-disable-next-line arrow-body-style
    return iterateCollection(
      c1,
      (c1Doc) => {
        return iterateCollection(`${c1}/${c1Doc.id}/${c2}`, handler && ((c2Doc) => handler(c2Doc, c1Doc)), c2opts).then(
          (ids) => {
            result[c1Doc.id] = ids;
          }
        );
      },
      c1opts
    ).return(result);
  };

  return {
    streamCollection,
    streamCollectionGroup,
    iterateSubCollection,
    iterateCollection,
    iterateCollectionGroup,
    subCollection: iterateSubCollection,
    collection: iterateCollection,
    collectionGroup: iterateCollectionGroup,
  };
};

module.exports = {
  genIteratorFuncs,
};
