const {
  UrlPattern,
  genMatcher,

  stringIsDoc,
  stringIsCollection,
  stringCollectionParts,
  urlPatternCollectionParts,
} = require('./match');

const DocumentTypes = Object.freeze({
  Assets: 'assets',
  AssetRefs: 'assetrefs',
  Boards: 'boards',
  BoardItems: 'boarditems',
  BoardActivities: 'boardactivities',
  InviteActions: 'inviteactions',
  Invites: 'invites',
  Offerers: 'offerers',
  Outclicks: 'outclicks',
  Profiles: 'profiles',
  ProfileActions: 'profileactions',
  ProfileFollowing: 'profilefollowing',
  ProfileFollowers: 'profilefollowers',
  ProfileLikes: 'profilelikes',
  ProfileFulfilled: 'profilefulfilled',
  ProfileDevices: 'profiledevices',
  ProfileNotifications: 'profilenotifications',
  ProfileInvites: 'profileinvites',
  ProfileSlugs: 'profile-slugs',
  UrlRefs: 'urlrefs',
  UrlRefsRefs: 'urlrefsrefs',
  UrlRefMap: 'urlrefmap',
  UrlRefScrapes: 'urlrefscrapes',
  Affiliates: 'affiliates',
  AffiliatePrograms: 'affiliateprograms',
  AffiliatePromotions: 'affiliatepromotions',
  AffiliateClicks: 'affiliateclicks',
  AffiliateFeeds: 'affiliatefeeds',
  Products: 'products',
});

const patterns = {
  [DocumentTypes.Assets]: {
    col: new UrlPattern('assets'),
    doc: new UrlPattern('assets/:publicid'),
  },
  [DocumentTypes.AssetRefs]: {
    col: new UrlPattern('assets/:assetId/refs'),
    doc: new UrlPattern('assets/:assetId/refs/:refId'),
  },
  [DocumentTypes.Boards]: {
    col: new UrlPattern('boards'),
    doc: new UrlPattern('boards/:boardid'),
  },
  [DocumentTypes.BoardItems]: {
    col: new UrlPattern('boards/:boardid/items'),
    doc: new UrlPattern('boards/:boardid/items/:itemid'),
    colgroup: 'items',
  },
  [DocumentTypes.BoardActivities]: {
    col: new UrlPattern('boards/:boardid/activities'),
    doc: new UrlPattern('boards/:boardid/activities/:activityid'),
    colgroup: 'activities',
  },
  [DocumentTypes.InviteActions]: {
    col: new UrlPattern(/^invites\/(.*?)\/actions$/, ['email']),
    doc: new UrlPattern(/^invites\/(.*?)\/actions\/(.*)$/, ['email', 'actionId']),
  },
  [DocumentTypes.Invites]: {
    col: new UrlPattern('invites'),
    doc: new UrlPattern(/^invites\/(.*)$/, ['email']),
  },
  [DocumentTypes.Offerers]: {
    col: new UrlPattern('offerers'),
    doc: new UrlPattern(/^offerers\/(.*)$/, ['domain']),
  },
  [DocumentTypes.Outclicks]: {
    col: new UrlPattern('offerers/:offerer/outclicks'),
    doc: new UrlPattern('offerers/:offerer/outclicks/:clickid'),
    colgroup: 'outclicks',
  },
  [DocumentTypes.Profiles]: {
    col: new UrlPattern('profiles'),
    doc: new UrlPattern('profiles/:uid'),
  },
  [DocumentTypes.ProfileActions]: {
    col: new UrlPattern('profiles/:uid/actions'),
    doc: new UrlPattern('profiles/:uid/actions/:actionId'),
  },
  [DocumentTypes.ProfileFollowing]: {
    col: new UrlPattern('profiles/:uid/following'),
    doc: new UrlPattern('profiles/:uid/following/:followUid'),
    colgroup: 'following',
  },
  [DocumentTypes.ProfileFollowers]: {
    col: new UrlPattern('profiles/:uid/followers'),
    doc: new UrlPattern('profiles/:uid/followers/:followUid'),
    colgroup: 'followers',
  },
  [DocumentTypes.ProfileLikes]: {
    col: new UrlPattern('profiles/:uid/likes'),
    doc: new UrlPattern('profiles/:uid/likes/:pathId'),
  },
  [DocumentTypes.ProfileFulfilled]: {
    col: new UrlPattern('profiles/:uid/fulfilled'),
    doc: new UrlPattern('profiles/:uid/fulfilled/:itemId'),
  },
  [DocumentTypes.ProfileDevices]: {
    col: new UrlPattern('profiles/:uid/devices'),
    doc: new UrlPattern('profiles/:uid/devices/:hardwareIdSha1'),
    colgroup: 'devices',
  },
  [DocumentTypes.ProfileNotifications]: {
    col: new UrlPattern('profiles/:uid/notifications'),
    doc: new UrlPattern('profiles/:uid/notifications/:notificationId'),
    colgroup: 'notifications',
  },
  [DocumentTypes.ProfileInvites]: {
    col: new UrlPattern('profiles/:uid/invites'),
    doc: new UrlPattern('profiles/:uid/invites/:inviteId'),
  },
  [DocumentTypes.ProfileSlugs]: {
    col: new UrlPattern('profile-slugs'),
    doc: new UrlPattern(/^profile-slugs\/(.*)$/, ['slug']),
  },
  [DocumentTypes.UrlRefs]: {
    col: new UrlPattern('urlrefs'),
    doc: new UrlPattern('urlrefs/:sha1'),
  },
  [DocumentTypes.UrlRefsRefs]: {
    col: new UrlPattern('urlrefs/:sha1/refs'),
    doc: new UrlPattern('urlrefs/:sha1/refs/:refId'),
  },
  [DocumentTypes.UrlRefMap]: {
    col: new UrlPattern('urlrefmap'),
    doc: new UrlPattern('urlrefmap/:sha1'),
  },
  [DocumentTypes.UrlRefScrapes]: {
    col: new UrlPattern('urlrefs/:sha1/scrapes'),
    doc: new UrlPattern('urlrefs/:sha1/scrapes/:scrapeId'),
  },
  [DocumentTypes.Affiliates]: {
    col: new UrlPattern('affiliates'),
    doc: new UrlPattern('affiliates/:affiliate'),
  },
  [DocumentTypes.AffiliatePrograms]: {
    col: new UrlPattern('affiliates/:affiliate/affiliate_programs'),
    doc: new UrlPattern('affiliates/:affiliate/affiliate_programs/:programId'),
  },
  [DocumentTypes.AffiliatePromotions]: {
    col: new UrlPattern('affiliates/:affiliate/affiliate_programs/:programId/affiliate_promotions'),
    doc: new UrlPattern('affiliates/:affiliate/affiliate_programs/:programId/affiliate_promotions/:promotionId'),
    colgroup: 'affiliate_promotions',
  },
  [DocumentTypes.AffiliateClicks]: {
    col: new UrlPattern('affiliates/:affiliate/affiliate_programs/:programId/affiliate_clicks'),
    doc: new UrlPattern('affiliates/:affiliate/affiliate_programs/:programId/affiliate_clicks/:clickId'),
  },
  [DocumentTypes.AffiliateFeeds]: {
    col: new UrlPattern('affiliates/:affiliate/affiliate_programs/:programId/affiliate_feeds'),
    doc: new UrlPattern('affiliates/:affiliate/affiliate_programs/:programId/affiliate_feeds/:feedId'),
  },
  [DocumentTypes.Products]: {
    col: new UrlPattern(
      'affiliates/:affiliate/affiliate_programs/:programId/affiliate_feeds/:feedId/affiliate_products'
    ),
    doc: new UrlPattern(
      'affiliates/:affiliate/affiliate_programs/:programId/affiliate_feeds/:feedId/affiliate_products/:productId'
    ),
    colgroup: 'affiliate_products',
  },
};

const {
  findDocType,
  findCollectionType,
  findCollectionGroup,
  getParamsByType,
  getDocPathByParams,
  getCollectionPathByParams,
} = genMatcher(patterns);

const getInfo = (refOrDocOrPath, fs) => {
  let path;
  let firestore;
  let snap;

  if (typeof refOrDocOrPath === 'string') {
    if (!fs) throw new Error('firestore needed as second arg when path provided as string');

    // if (!stringIsDoc(refOrDocOrPath))
    //     throw new Error(`not a valid doc path: ${refOrDocOrPath}`);

    firestore = fs;
    path = refOrDocOrPath;
  }

  if (!path && refOrDocOrPath.path) {
    firestore = refOrDocOrPath.firestore;
    path = refOrDocOrPath.path;
  }

  if (!path && refOrDocOrPath.ref) {
    firestore = refOrDocOrPath.ref.firestore;
    path = refOrDocOrPath.ref.path;
    if (typeof refOrDocOrPath.data === 'function') snap = refOrDocOrPath;
  }

  if (!path) throw new Error('invalid path or input');

  let type;
  let isCollection;

  type = findDocType(path);

  if (!type) {
    type = findCollectionType(path);
    if (type) isCollection = true;
  }

  if (!type) throw new Error(`unknown path: ${path}`);

  return {
    type,
    ref: isCollection ? firestore.collection(path) : firestore.doc(path),
    snap,
    isCollection: !!isCollection,
  };
};

module.exports = {
  patterns,
  findDocType,
  findCollectionType,
  findCollectionGroup,
  getParamsByType,
  getDocPathByParams,
  getCollectionPathByParams,
  DocumentTypes,
  stringIsDoc,
  stringIsCollection,
  stringCollectionParts,
  urlPatternCollectionParts,
  getInfo,
};
