const test = require('tape');

const {
  findDocType,
  findCollectionType,
  findCollectionGroup,
  getParamsByType,
  getInfo,
  getDocPathByParams,
  getCollectionPathByParams,
} = require('./index');

const { createSnap } = require('../test')();

test('findDocType', (t) => {
  t.deepEqual(findDocType('boards/1'), 'boards');
  t.deepEqual(findDocType('boards/1/items/1'), 'boarditems');
  t.deepEqual(findDocType('offerers/123inkt.nl'), 'offerers');
  t.deepEqual(findDocType('invites/floris.vink@ooshopping.nl'), 'invites');
  t.deepEqual(findDocType('invites/japie@dingen.nl/actions/123'), 'inviteactions');
  t.deepEqual(findDocType('profile-slugs/hoi-doei-dag'), 'profile-slugs');
  t.deepEqual(findDocType('profile-slugs/e.j.vd.hooven'), 'profile-slugs');
  t.deepEqual(findDocType('urlrefmap/0003dd7e00642292af2178d22a7652eea7d324d4'), 'urlrefmap');
  t.deepEqual(findDocType('profiles/0QHx6kQwkyWtJnrK42pk7wsxTd62/invites/VajzBacg47Gplbi6hlC0'), 'profileinvites');
  t.deepEqual(
    findDocType('affiliates/awin/affiliate_programs/10011/affiliate_promotions/782117'),
    'affiliatepromotions'
  );

  t.end();
});

test('findCollectionType', (t) => {
  t.deepEqual(findCollectionType('boards'), 'boards');
  t.deepEqual(findCollectionType('invites'), 'invites');
  t.deepEqual(findCollectionType('invites/howdy@ho.nl/actions'), 'inviteactions');
  t.deepEqual(findCollectionType('boards/1/items'), 'boarditems');
  t.deepEqual(findCollectionType('profiles/0QHx6kQwkyWtJnrK42pk7wsxTd62/invites'), 'profileinvites');
  t.deepEqual(findCollectionType('urlrefmap'), 'urlrefmap');
  t.deepEqual(findCollectionType('boards/1/items/1'), undefined);

  t.end();
});

test('findCollectionGroup', (t) => {
  t.deepEqual(findCollectionGroup('foobar'), undefined);
  t.deepEqual(findCollectionGroup('items'), 'boarditems');
  t.deepEqual(findCollectionGroup('activities'), 'boardactivities');

  t.end();
});

test('getParamsByType', (t) => {
  t.deepEqual(getParamsByType('boards/123/items/456'), {
    type: 'doc',
    docType: 'boarditems',
    params: { boardid: '123', itemid: '456' },
  });
  t.deepEqual(getParamsByType('boards/123'), { type: 'doc', docType: 'boards', params: { boardid: '123' } });
  t.deepEqual(getParamsByType('boards'), { type: 'col', docType: 'boards', params: {} });
  t.deepEqual(getParamsByType('foobar'), undefined);

  t.end();
});

test('getInfo by path+firestore', (t) => {
  const path = 'boards/123/items/456';
  const snap = createSnap({}, path);
  const result = getInfo(path, snap.ref.firestore);

  t.equal(result.type, 'boarditems');
  t.equal(result.ref.path, path);

  t.end();
});

test('getInfo by docref', (t) => {
  const path = 'boards/123/items/456';
  const snap = createSnap({}, path);
  const result = getInfo(snap.ref);

  t.equal(result.type, 'boarditems');
  t.equal(result.ref.path, path);

  t.end();
});

test('getInfo by snap', (t) => {
  const path = 'boards/123/items/456';
  const snap = createSnap({ a: 2 }, path);
  const result = getInfo(snap);

  t.equal(result.type, 'boarditems');
  t.equal(result.ref.path, path);
  t.equal(result.snap.get('a'), 2);

  t.end();
});

test('getInfo by col ref', (t) => {
  const path = 'boards/123/items/456';
  const colRef = createSnap({ a: 2 }, path).ref.parent;
  const result = getInfo(colRef);

  t.equal(result.type, 'boarditems');
  t.equal(result.ref.path, 'boards/123/items');
  t.equal(result.isCollection, true);

  t.end();
});

test('getInfo by col path', (t) => {
  const path = 'boards/123/items';
  // eslint-disable-next-line prefer-destructuring
  const firestore = createSnap().ref.firestore;
  const result = getInfo(path, firestore);

  t.equal(result.type, 'boarditems');
  t.equal(result.ref.path, 'boards/123/items');
  t.equal(result.isCollection, true);

  t.end();
});

test('getDocPathByParams', (t) => {
  const boardid = 'board1';
  const itemid = 'item1';

  t.throws(() => getDocPathByParams('foobar'));
  t.throws(() => getDocPathByParams('boarditems', { foobar: 1 }));
  t.deepEqual(getDocPathByParams('boarditems', { boardid, itemid }), `boards/${boardid}/items/${itemid}`);

  t.end();
});

test('getCollectionPathByParams', (t) => {
  const boardid = 'board1';

  t.throws(() => getCollectionPathByParams('foobar'));
  t.throws(() => getCollectionPathByParams('boarditems', { foobar: 1 }));
  t.deepEqual(getCollectionPathByParams('boarditems', { boardid }), `boards/${boardid}/items`);

  t.end();
});
