const P = require('bluebird');

const differenceBy = require('lodash/differenceBy');
const omit = require('lodash/omit');
const partialRight = require('lodash/partialRight');
const pick = require('lodash/pick');
const noop = require('lodash/noop');

const diff = require('jsondiffpatch');

const { notThrowIfNotFound } = require('../errors');

const { toPath, collectRefs, refToPath } = require('../');
const { findDocType } = require('../match');

const { stripNonPojo } = require('@hatsa/utils/object');

//created a docPath to refDoc in a ref sub-collection of the document
const refDoc = (doc, ref) => ref.firestore.doc(`${toPath(doc)}/refs/${ref.path.split('/').join('_')}`);

const addRef = (docPath, ref, transaction) => {
  const doc = refDoc(docPath, ref);
  const data = {
    ref,
    refPath: toPath(ref),
  };

  if (!transaction) return doc.set(data);

  return transaction.set(doc, data);
};

const removeRef = (docPath, ref, transaction) => {
  const doc = refDoc(docPath, ref);

  if (!transaction) return doc.delete();

  return transaction.delete(doc);
};

const updateSubColCount = (doc, subColPath, countProp) =>
  P.resolve(doc.collection(subColPath).get())
    .then((ss) => ss.size)
    .tap((count) => console.log(`updateSubColCount: setting ${countProp} for ${doc.path} to ${count}`)) //eslint-disable-line
    .tap((count) => doc.update({ [countProp]: count /*, [UPDATED_AT_PROP]: timestamp*/ }))
    .catch(notThrowIfNotFound);
const updateRefCount = partialRight(updateSubColCount, 'refs', 'refcount');

const getRefsDelta = (change, opts = {}) => {
  const { omitPaths, alwaysAdd } = opts;

  const newDoc = omit(change.after.data() || {}, omitPaths || []);
  const oldDoc = omit(change.before.data() || {}, omitPaths || []);

  const oldrefs = collectRefs(oldDoc);
  const newrefs = collectRefs(newDoc);

  const refsToBeDeleted = differenceBy(oldrefs, newrefs, toPath);
  const refsToBeAdded = alwaysAdd !== false ? newrefs : differenceBy(newrefs, oldrefs, toPath);

  return {
    refsToBeDeleted,
    refsToBeAdded,
    oldrefs,
    newrefs,
    newDoc,
    oldDoc,
  };
};

/**
 * Traverses the document and find all References. On all the target documents a ref-document will be added in the /refs/ colleciton.
 * Old ref-doucments will be removed
 */
const setRefs = (change, opts = {}) => {
  const firestore = change.before.ref.firestore;

  const { omitPaths, alwaysAdd, logger, debug } = opts;

  const log = logger ? logger.info : noop;

  const doc = change.after.ref;

  const { refsToBeDeleted, refsToBeAdded, oldrefs, newrefs, newDoc, oldDoc } = getRefsDelta(change, {
    omitPaths,
    alwaysAdd,
  });

  if (debug)
    log('setRefs', {
      docs: {
        newDoc,
        oldDoc,
      },
      refs: {
        oldrefs,
        newrefs,
      },
      refsToHandle: {
        refsToBeDeleted,
        refsToBeAdded,
      },
    });

  if (!refsToBeDeleted.length && !refsToBeAdded.length) return Promise.resolve();

  if (refsToBeAdded.length) log(`adding refs from ${toPath(doc)} to `, refsToBeAdded.map(toPath).join(', '));

  if (refsToBeDeleted.length) log(`removing refs from ${toPath(doc)} to `, refsToBeDeleted.map(toPath).join(', '));

  const batch = firestore.batch();

  refsToBeDeleted.forEach((ref) => removeRef(ref, doc, batch));
  refsToBeAdded.forEach((ref) => addRef(ref, doc, batch));

  return batch.commit();
};

/**
 * Will look for a subcollection "refs" on the document and updates all those documents (linked by: <Reference>:ref property)
 * Optionally a updateFunc can be given to change this behaviour (src-ref-doc, target-ref-doc) => {...}
 * If target-doc does not exists it will be skipped
 */
const populateRefs = (change, updateFunc, opts = {}) => {
  const { logger } = opts;

  const log = logger ? logger.info : console.log; //eslint-disable-line

  return change.after.ref
    .collection('refs')
    .get()
    .then((ss) => {
      if (!ss.docs.length) return Promise.resolve();

      const sourceRefPath = change.after.ref.path;

      log(
        `populateRefs: populating ${ss.docs.length} docs from ${sourceRefPath}`,
        ss.docs.map((d) => d.id)
      ); //eslint-disable-line

      const populateProm = (refDoc) => {
        const targetRef = refDoc.get('ref');
        const targetRefPath = refToPath(targetRef);

        if (!targetRef) return Promise.resolve();

        if (typeof updateFunc !== 'function') {
          //Automatic - the entire document will be embedded in the target document

          const updateField = findDocType(sourceRefPath) || sourceRefPath;
          return targetRef.update({ [updateField]: change.after.data() }).catch(notThrowIfNotFound);
        }

        //Manual - user has control what will be updates. updateFunc signature: (refToDocThatHasBeenChanged, refTotargetDoc)
        return targetRef
          .get()
          .then((doc) => updateFunc(change.after, doc))
          .then((updateMap) => {
            log(`populateRefs: saving from ${sourceRefPath} to ${targetRefPath}`); //eslint-disable-line

            if (Array.isArray(updateMap)) {
              return Promise.all(
                updateMap.map(({ fieldPath, fieldValue }) => {
                  log(`populateRefs: field > value: ${fieldPath}`, fieldValue); //eslint-disable-line
                  return targetRef.update(fieldPath, fieldValue).catch(notThrowIfNotFound);
                })
              );
            }

            const updatedFields = Object.keys(updateMap || {});

            if (!updatedFields.length) return Promise.resolve();

            const delta = diff.diff(stripNonPojo(pick(change.after.data(), updatedFields)), updateMap);
            log('populateRefs: updateMap: ', delta); //eslint-disable-line

            return targetRef.update(updateMap).catch(notThrowIfNotFound);
          });
      };

      return P.map(ss.docs, populateProm, { concurrency: 8 });
    });
};

module.exports = {
  //interal, expose for testing
  getRefsDelta,
  refDoc,

  addRef,
  removeRef,
  setRefs,
  populateRefs,
  updateRefCount,
  updateSubColCount,
};
