const Ajv = require('ajv');
const formats = require('ajv/lib/compile/formats')('fast');

const path = require('path');
const castArray = require('lodash/castArray');

const { findDocType } = require('../match');
const loadJsFiles = require('./loadJsFiles');

const augmentAjv = (ajv) => {
  const classes = { Date };
  ajv.addKeyword('instanceOf', {
    compile: (cls) => (data) => data instanceof classes[cls],
  });

  ajv.addKeyword('firestoreReference', {
    compile: (docTypes) => (data) => castArray(docTypes).some((docType) => findDocType(data.path) === docType),
  });
  ajv.addKeyword('firestoreTimestamp', {
    compile: () => (data) => data && typeof data.toDate === 'function',
  });

  ajv.addFormat('url_rel_proto', (d) => {
    const url = encodeURI(d);
    return formats.url.test(url) || formats.url.test(`https:${url}`);
  });
  ajv.addFormat('domain', (d) => /\b((?=[a-z0-9-]{1,63}\.)(xn--)?[a-z0-9]+(-[a-z0-9]+)*\.)+[a-z]{2,63}\b/.test(d));
};

const getSchemaFromModule = (schemaModule) => {
  // If schema module contains export like `*Schema`, e.g. `InviteSchema`
  // we assume that it's our schema object, otherwise schema is exported as default export.
  const keys = Object.keys(schemaModule);
  const schemaKey = keys.find((key) => new RegExp('.*Schema$').test(key));
  const isValidSchema = (s) => s && typeof s === 'object' && Object.prototype.hasOwnProperty.call(s, '$id');

  if (isValidSchema(schemaModule[schemaKey])) {
    return schemaModule[schemaKey];
  }

  if (isValidSchema(schemaModule)) {
    return schemaModule;
  }

  return null;
};

const getAjv = (preloadSchemas = true, schemas = []) => {
  const preloadedSchemas = [];

  if (preloadSchemas) {
    preloadedSchemas.push(
      ...loadJsFiles(path.join(__dirname, '/../schemas/firestore')).map(getSchemaFromModule).filter(Boolean)
    );
    preloadedSchemas.push(
      ...loadJsFiles(path.join(__dirname, '/../schemas/elasticsearch')).map(getSchemaFromModule).filter(Boolean)
    );
  }

  if (schemas.length > 0) {
    preloadedSchemas.push(...schemas);
  }

  const ajv = new Ajv({
    verbose: true,
    missingRefs: true,
    extendRefs: true,
    schemas: preloadedSchemas,
  });

  augmentAjv(ajv);

  return ajv;
};

const getSchemaId = (s) => `http://hopponn.com/schemas/${s}.json`;

module.exports = {
  getAjv,
  getSchemaId,
};
