import * as ajv from 'ajv';
import { HatsaLogger } from '@hatsa/log';
import { DocumentSnapshot } from 'firebase-admin'; // eslint-disable-line import/no-extraneous-dependencies
import { Schema } from '../schemas/schema';

declare namespace Validate {
  type ValidateOpts = {
    logger?: HatsaLogger;
    doThrow?: boolean;
    debug?: boolean;
  };

  export type Ajv = ajv.Ajv;

  /**
   * Validation error is thrown when validation fails and doThrow option is set to true.
   */
  export class SchemaValidationError extends Error {
    code: 'SCHEMA_VALIDATION_ERROR';

    schema: string;

    schemaErrors: ajv.ErrorObject[];

    data: Record<string, unknown>;
  }

  interface HatsaValidator {
    /**
     * Validator instance.
     */
    validatorInstance: ajv.Ajv;

    /**
     * Validate object. Make sure schema is preloaded before.
     *
     * @param {object} data - The object to validate
     * @param {string | Schema} schema - The schema id.
     * @param {ValidateOpts} opts - Validator options.
     */
    validate<T>(data: unknown, schema: string | Schema, opts?: ValidateOpts): data is T;

    /**
     * Validate firebase document. Before using, make sure all schemas were preloaded before.
     *
     * @param {DocumentSnapshot} doc - Firebase document snapshot.
     * @param {ValidateOpts} opts - Validator options.
     */
    validateDoc(doc: DocumentSnapshot, opts: ValidateOpts): boolean;

    /**
     * Schema validation error class.
     */
    SchemaValidationError: SchemaValidationError;
  }

  /**
   * Generate validator.
   * @param {boolean} preloadSchemas - indicates whether we need to preload all schemas from "@hatsa/firestore" package.
   * @param {Array.<Schema>} schemas - Instantiate with predefined schemas.
   */
  export function genValidator(preloadSchemas?: boolean, schemas?: Schema[]): HatsaValidator;

  /**
   * Convert validator errors to a string.
   * @param errors
   */
  export function stringifyErrors(errors: ajv.ErrorObject[]): string;
}

export = Validate;
