const omit = require('lodash/omit');
const ajvErrors = require('better-ajv-errors');

const { refToPath } = require('../index');
const { findDocType } = require('../match');

const { getAjv, getSchemaId } = require('./ajv');

class SchemaValidationError extends Error {
  constructor(schema, schemaErrors, data) {
    super(`Schema validation error on: ${schema}`);
    this.code = 'SCHEMA_VALIDATION_ERROR';
    this.schema = schema;
    this.schemaErrors = schemaErrors;
    this.data = data;
    this.beautifulAjvErrors = ajvErrors(schema, data, schemaErrors, { indent: 2 });
  }
}

/*
const logSchemaErrorMessage = (errs, logger) => {

    const log = logger ? logger.debug : console.log; //eslint-disable-line

    errs.forEach((err, idx) => {

        let errorline = `[${idx}]   error: "${err.dataPath || '.'}" ${err.message}`;

        switch (err.keyword) {
            case 'additionalProperties':
                errorline += ` "${err.params.additionalProperty}"`;
                break;
        }

        // log(JSON.stringify(err));
        log(errorline);

        const debugInfo = omit(err, 'data', 'schema', 'parentSchema');
        log(`keyword    : ${debugInfo.keyword}`);
        log(`dataPath   : ${debugInfo.dataPath}`);
        log(`schemaPath : ${debugInfo.schemaPath}`);
        log(`params     : ${JSON.stringify(debugInfo.params)}`);
        log(`message    : ${debugInfo.message}`);
        // log(`[${idx}] given: (${typeof err.data}): ${JSON.stringify(stripNonPojo(err.data, v => `<<${getOCName(v)}>>`), null, 2)}`);
        // log("Doc:");
        // jsbprint(d);
    });
};
*/

const stringifyErrors = (errs) => {
  if (!Array.isArray(errs)) return [];

  return errs
    .map((err, idx) => {
      let errorline = `[${idx}]   error: "${err.dataPath || '.'}" ${err.message}`;

      switch (err.keyword) {
        case 'additionalProperties':
          errorline += ` "${err.params.additionalProperty}"\n`;
          break;
        default:
          break;
      }

      const debugInfo = omit(err, 'data', 'schema', 'parentSchema');
      errorline += `keyword    : ${debugInfo.keyword}\n`;
      errorline += `dataPath   : ${debugInfo.dataPath}\n`;
      errorline += `schemaPath : ${debugInfo.schemaPath}\n`;
      errorline += `params     : ${JSON.stringify(debugInfo.params)}\n`;
      errorline += `message    : ${debugInfo.message}\n`;

      return errorline;

      // log(JSON.stringify(err));
      // log(errorline);

      // const debugInfo = omit(err, 'data', 'schema', 'parentSchema');
      // log(`keyword    : ${debugInfo.keyword}`);
      // log(`dataPath   : ${debugInfo.dataPath}`);
      // log(`schemaPath : ${debugInfo.schemaPath}`);
      // log(`params     : ${JSON.stringify(debugInfo.params)}`);
      // log(`message    : ${debugInfo.message}`);
      // log(`[${idx}] given: (${typeof err.data}): ${JSON.stringify(stripNonPojo(err.data, v => `<<${getOCName(v)}>>`), null, 2)}`);
      // log("Doc:");
      // jsbprint(d);
    })
    .join('\n');
};

const genValidator = (preloadSchemas = true, schemas = []) => {
  const ajvInst = getAjv(preloadSchemas, schemas); // does synchronous loading of schema definition files

  const hasSchemaErrors = (schema, data) => {
    const v = ajvInst.validate(getSchemaId(schema), data);
    return !v ? ajvInst.errors : false;
  };

  /**
   * Get schema name from schema id,
   *
   * @param {string} schemaId - Id of the schema.
   * @returns {string}
   */
  const getSchemaName = (schemaId) => {
    const [, schemaName] = new RegExp('http://hopponn.com/schemas/(.*?).json').exec(schemaId) || [];
    return schemaName;
  };

  /**
   * Add schema to the validator.
   *
   * @param {*} schema - Schema
   */
  const addSchema = (schema) => {
    if (!ajvInst.getSchema(schema.$id)) {
      ajvInst.addSchema(schema);
    }
  };

  const validate = (data, schema, opts = {}) => {
    const schemaName = typeof schema === 'object' ? getSchemaName(schema.$id) : schema;
    const { doThrow, debug, logger } = opts;

    // When schema object is used, add it to the validator instance,
    // so we can validate based this schema.
    if (typeof schema === 'object' && schema.$id && schema.$id.length > 0) {
      addSchema(schema);
    }

    const errors = hasSchemaErrors(schemaName, data);

    if (!errors || !errors.length) return true;

    if (doThrow) throw new SchemaValidationError(schemaName, errors, data);

    if (debug && logger) logger.debug(stringifyErrors(errors));

    if (debug && !logger) {
      console.log(ajvErrors(schema, data, errors, { indent: 2 })); //eslint-disable-line
      // console.log(stringifyErrors(errors)); //eslint-disable-line
    }

    // console.log(`given: ${JSON.stringify(stripNonPojo(d, v => `<<${getOCName(v)}>>`), null, 2)}`); //eslint-disable-line
    // if (debug)
    //     logSchemaErrorMessage(errors, opts.logger);

    return false;
  };

  const validateDoc = (doc, opts = {}) => {
    // const log = opts.logger ? opts.logger.info : console.log; //eslint-disable-line
    const { useEsSchema } = opts;
    const docPath = refToPath(doc.ref);
    const docSchemaType = findDocType(docPath);

    const data = doc.data();

    // if (opts.debug)
    //     log(`validating ${docSchemaType}: ${doc.ref.firestore.projectId}/${docPath}`);
    const schemaSuffix = useEsSchema ? 'DocEs' : 'Doc';
    return validate(data, `${docSchemaType}${schemaSuffix}`, opts);
  };

  return {
    validatorInstance: ajvInst,
    validateDoc,
    validate,
    addSchema,
    // logSchemaErrorMessage,
    SchemaValidationError,
  };
};

module.exports = {
  // logSchemaErrorMessage,
  genValidator,
  SchemaValidationError,
  stringifyErrors,
};
